import numpy as np
from skimage.exposure import match_histograms


def match_hist_all(all_ds_avg_ch):
    ref = all_ds_avg_ch[0][0]
    all_ds_avg_ch_matched = []
    for ds_avg_ch in all_ds_avg_ch:
        ds_avg_ch_matched = []
        for i in range(len(ds_avg_ch)):
            ds_avg_ch_matched.append(match_histograms(ds_avg_ch[i], ref))
        all_ds_avg_ch_matched.append(ds_avg_ch_matched)
        
    return all_ds_avg_ch_matched

def make_rgb_img(img1, img2):
    img_rgb = np.zeros((img1.shape[0], img1.shape[1], 3))
    # normalise images to 0-1
    img1_norm = (img1 - np.min(img1)) / (np.max(img1) - np.min(img1))
    img2_norm = (img2 - np.min(img2)) / (np.max(img2) - np.min(img2))
    img_rgb[:, :, 0] = img1_norm
    img_rgb[:, :, 1] = img2_norm
    
    return img_rgb

def saturate_perc(img_rgb, sat_perc=99):
    img_rgb = np.clip(img_rgb, 0, np.percentile(img_rgb, sat_perc)) 
    img_rgb = (img_rgb / np.max(img_rgb) * 255).astype(np.uint8)
    return img_rgb

def get_all_wind_mean_img(all_ds_mean_img, all_ds_centroids, all_pl_match_mat, nrn_id, plane_idx=0, win_size=64):

    all_wind_mean_img = []
    for i in range(len(all_ds_mean_img)):
        centroid_ids = all_pl_match_mat[plane_idx][nrn_id,:]
        cent = all_ds_centroids[i][plane_idx][centroid_ids[i]]
        mean_img = all_ds_mean_img[i][plane_idx]

        # pad mean image with mean and shift the centroid appropriately
        # mean_img = np.pad(mean_img, ((win_size,win_size),(win_size,win_size)))
        mean_img = np.pad(mean_img, ((win_size,win_size),(win_size,win_size)), mode='constant', constant_values=0)
        cent = cent + win_size
    
        wind_mean_img = mean_img[int(cent[0]-win_size/2):int(cent[0]+win_size/2), int(cent[1]-win_size/2):int(cent[1]+win_size/2)]


        all_wind_mean_img.append(wind_mean_img)

    return all_wind_mean_img


import numpy as np

def load_track_ops(track_ops_path):
    track_ops = np.load(track_ops_path + 'track_ops_postreg.npy', allow_pickle=True).item()
    return track_ops

def load_stat_ds_plane(track_ops_path, track_ops, plane_idx=0):
    stat = np.load(track_ops_path + f'/suite2p/plane{plane_idx}/stat.npy', allow_pickle=True)
    iscell = np.load(track_ops_path + f'/suite2p/plane{plane_idx}/iscell.npy', allow_pickle=True)
    # filter based on track_ops.iscell_thr
    len_stat_allcell = len(stat)
    
    if track_ops.iscell_thr==None:
        stat = stat[iscell[:,0] ==1]
    else:
        stat= stat[iscell[:,1]>track_ops.iscell_thr]
    len_stat_iscell = len(stat)
    print(f'Loading ROIs for plane{plane_idx} in dataset {track_ops_path.split("/")[-2]}')
    print(f'Chose {len_stat_iscell}/{len_stat_allcell} ROIs, based on s2p iscell threshold {track_ops.iscell_thr} (see track_ops.iscell_thr)')
    # make a stat_summary dictionary
    stat_summary = {
        'len_stat_allcell': len_stat_allcell,
        'len_stat_iscell': len_stat_iscell,
        'iscell_thr': track_ops.iscell_thr
    }
    return stat, stat_summary

def get_all_roi_array_from_stat(stat, track_ops):
    n_xpix = track_ops.all_ds_avg_ch1[0][0].shape[0]
    n_ypix = track_ops.all_ds_avg_ch1[0][0].shape[0]
    all_roi_array = np.zeros((n_xpix, n_ypix, len(stat)), bool)

    for i in range(len(stat)):

        roi_xpix = stat[i]['xpix']
        roi_ypix = stat[i]['ypix']
     
        # Convert the ROI coordinates into a grid of values
        roi_grid = np.zeros((n_xpix, n_ypix), dtype=np.float32)
        roi_grid[np.array(roi_ypix), np.array(roi_xpix)] = 1

        all_roi_array[:,:,i] = roi_grid
        
    return all_roi_array




from track2p.ops.default import DefaultTrackOps
from types import SimpleNamespace

from track2p.io.s2p_loaders import load_all_imgs, check_nplanes, load_all_ds_stat_iscell, load_all_ds_mean_img, load_all_ds_centroids
from track2p.io.savers import save_track_ops, save_all_pl_match_mat

from track2p.register.loop import run_reg_loop, reg_all_ds_all_roi
from track2p.register.utils import get_all_ds_img_for_reg, get_all_ref_nonref_inters

from track2p.plot.progress import plot_all_planes
from track2p.plot.output import plot_reg_img_output, plot_thr_met_hist, plot_n_matched_roi, plot_roi_reg_output, plot_roi_match_multiplane, plot_allroi_match_multiplane

from track2p.match.loop import get_all_ds_assign, get_all_pl_match_mat 
import numpy as np
import os


def run_t2p(track_ops):

    # 1) initialise save paths for figures and matched neurons output
    track_ops.init_save_paths()

    # 2) Load data
    check_nplanes(track_ops)
    all_ds_avg_ch1, all_ds_avg_ch2 = load_all_imgs(track_ops)

    # 3) Plot available planes for registration
    plot_all_planes(all_ds_avg_ch1, track_ops)
    if track_ops.nchannels==2:
        plot_all_planes(all_ds_avg_ch2, track_ops, ch='anatomical')

    # 4) do the actual registration based on chosen channel
    all_ds_ref_img, all_ds_mov_img = get_all_ds_img_for_reg(all_ds_avg_ch1, all_ds_avg_ch2, track_ops)

    all_ds_mov_img_reg, all_ds_reg_params = run_reg_loop(all_ds_ref_img, all_ds_mov_img, track_ops) # TODO: save basic parameters for each registration as feedback (e. g. ammoung of shift, rotation, etc.) for later plotting

    plot_reg_img_output(track_ops)
    

    # 5) apply computed transorm to all ROIs
    all_ds_all_roi_ref, all_ds_all_roi_mov, all_ds_all_roi_reg, all_ds_roi_counter = reg_all_ds_all_roi(all_ds_reg_params, track_ops)
 

    # 6) optional: generate 'yellow intersection' plot (this is only needed for plotting below)
    all_ds_ref_reg_inters = get_all_ref_nonref_inters(all_ds_all_roi_ref, all_ds_all_roi_reg, track_ops)

    all_ds_ref_mov_inters = get_all_ref_nonref_inters(all_ds_all_roi_ref, all_ds_all_roi_mov, track_ops)


    track_ops.all_ds_ref_mov_inters = all_ds_ref_mov_inters
    track_ops.all_ds_ref_reg_inters = all_ds_ref_reg_inters

    # this line is very memory-intensive because of the ROIS (TODO: maybe instead of contours just plot RGB) (or somehow generate RGB image of contours (in the part before))
    if track_ops.show_roi_reg_output:
        plot_roi_reg_output(track_ops)
      

    # 7) get optimal assignments for all pairs of recordings (first to last)
    all_ds_assign, all_ds_assign_thr, all_ds_thr_met, all_ds_thr = get_all_ds_assign(track_ops, all_ds_all_roi_ref, all_ds_all_roi_reg)
    plot_thr_met_hist(all_ds_thr_met, all_ds_thr, track_ops)
    plot_n_matched_roi(all_ds_thr_met, all_ds_thr, track_ops)


    # 8) get match matrices for all pairs of recordings (first to last)
    all_pl_match_mat = get_all_pl_match_mat(all_ds_all_roi_ref, all_ds_assign_thr, track_ops)


    # 9) save results
    save_track_ops(track_ops)

    
    save_all_pl_match_mat(all_pl_match_mat, track_ops)


    # 10) save in suite2p format
    if track_ops.save_in_s2p_format:
        print('Saving in suite2p format...')
        save_in_s2p_format(track_ops)
        
    # 10) plot results
    print('Generating plots (this can take some time)...')
    all_ds_stat_iscell = load_all_ds_stat_iscell(track_ops)
    all_ds_centroids = load_all_ds_centroids(all_ds_stat_iscell, track_ops)
    all_ds_mean_img = load_all_ds_mean_img(track_ops)
    if track_ops.nchannels==2:
        all_ds_mean_img_ch2 = load_all_ds_mean_img(track_ops, ch=2)


    plot_roi_match_multiplane(all_ds_mean_img, all_ds_centroids, all_pl_match_mat, track_ops, win_size=track_ops.win_size) # TODO: match histogram to the first roi of first batch (not first roi of each batch)
    plot_allroi_match_multiplane(all_ds_mean_img, all_pl_match_mat, track_ops)
    if track_ops.nchannels==2:
        plot_roi_match_multiplane(all_ds_mean_img_ch2, all_ds_centroids, all_pl_match_mat, track_ops, win_size=track_ops.win_size, ch=2) # TODO: match histogram to the first roi of first batch (not first roi of each batch)
        plot_allroi_match_multiplane(all_ds_mean_img_ch2, all_pl_match_mat, track_ops, ch=2)


    
    print('\n\n\nDone!\n\n\n')
    

def save_in_s2p_format(track_ops):
    
    for ds_path in track_ops.all_ds_path:
        # check how many subfolders starting with plane* in suite2p folder
        n_planes = len([name for name in os.listdir(ds_path + '/suite2p') if name.startswith('plane')])
        print(f'Found {n_planes} planes in {ds_path}')

    folderpath=track_ops.save_path
    track_ops_dict = np.load(os.path.join(folderpath,  "track_ops.npy"), allow_pickle=True).item()
    track_ops = SimpleNamespace(**track_ops_dict)
    iscell_thr = track_ops.iscell_thr


    for j in range(track_ops.nplanes):
        
        all_f_t2p= []
        all_ops = []
        all_stat_t2p = []
        all_iscell_t2p = []
        fneu_iscell_t2p= []
        spks_iscell_t2p= []
        
        fneu_chan2_iscell_t2p = []
        f_chan2_iscell_t2p = []
        redcell_iscell_t2p = []
            
        
        print(f'nplanes{j}')
        t2p_match_mat = np.load(os.path.join(folderpath,f'plane{str(j)}_match_mat.npy'), allow_pickle=True)
        t2p_match_mat_allday = t2p_match_mat[~np.any(t2p_match_mat == None, axis=1), :]
        for (i, ds_path) in enumerate(track_ops.all_ds_path):
                    ops = np.load(os.path.join(ds_path, 'suite2p', f'plane{str(j)}', 'ops.npy'), allow_pickle=True).item()
                    stat = np.load(os.path.join(ds_path, 'suite2p', f'plane{str(j)}', 'stat.npy'), allow_pickle=True)
                    f = np.load(os.path.join(ds_path, 'suite2p', f'plane{str(j)}', 'F.npy'), allow_pickle=True)
                    fneu= np.load(os.path.join(ds_path, 'suite2p', f'plane{str(j)}', 'Fneu.npy'), allow_pickle=True)
                    spks= np.load(os.path.join(ds_path, 'suite2p', f'plane{str(j)}', 'spks.npy'), allow_pickle=True)
                    iscell = np.load(os.path.join(ds_path, 'suite2p', f'plane{str(j)}', 'iscell.npy'), allow_pickle=True)
                    if track_ops.nchannels==2:
                        f_chan2=np.load(os.path.join(ds_path, 'suite2p', f'plane{str(j)}', 'F_chan2.npy'), allow_pickle=True)
                        fneu_chan2 = np.load(os.path.join(ds_path, 'suite2p', f'plane{str(j)}', 'Fneu_chan2.npy'), allow_pickle=True)
                        redcell=np.load(os.path.join(ds_path, 'suite2p', f'plane{str(j)}', 'redcell.npy'), allow_pickle=True)

                    if track_ops.iscell_thr==None:
                        stat_iscell = stat[iscell[:, 0] == 1]
                        f_iscell = f[iscell[:, 0] == 1, :]
                        fneu_iscell = fneu[iscell[:, 0] == 1, :]
                        spks_iscell = spks[iscell[:, 0] == 1, :]
                        is_cell = iscell[iscell[:, 0] == 1, :]
                        if track_ops.nchannels==2:
                            f_chan2_iscell = f_chan2[iscell[:, 0] == 1, :]
                            fneu_chan2_iscell = fneu_chan2[iscell[:, 0] == 1, :]
                            redcell_iscell = redcell[iscell[:, 0] == 1]
                    else:
                        stat_iscell = stat[iscell[:, 1] > iscell_thr]
                        f_iscell = f[iscell[:, 1] > iscell_thr, :]
                        fneu_iscell = fneu[iscell[:, 1] > iscell_thr, :]
                        spks_iscell = spks[iscell[:, 1] > iscell_thr, :]
                        is_cell = iscell[iscell[:, 1] > iscell_thr, :]
                        if track_ops.nchannels==2:
                            f_chan2_iscell = f_chan2[iscell[:, 1] > track_ops.iscell_thr, :]
                            fneu_chan2_iscell = fneu_chan2[iscell[:, 1] > track_ops.iscell_thr, :]
                            redcell_iscell = redcell[iscell[:, 1] > track_ops.iscell_thr]
                    
                    stat_t2p = stat_iscell[t2p_match_mat_allday[:, i].astype(int)]
                    f_t2p = f_iscell[t2p_match_mat_allday[:, i].astype(int), :]
                    fneu_t2p = fneu_iscell[t2p_match_mat_allday[:, i].astype(int), :]
                    spks_t2p = spks_iscell[t2p_match_mat_allday[:, i].astype(int), :]
                    iscell_t2p = is_cell[t2p_match_mat_allday[:, i].astype(int), :]
                    if track_ops.nchannels==2:
                        fneu_chan2_t2p = fneu_chan2_iscell[t2p_match_mat_allday[:, i].astype(int), :]
                        f_chan2_t2p = f_chan2_iscell[t2p_match_mat_allday[:, i].astype(int), :]
                        redcell_t2p = redcell_iscell[t2p_match_mat_allday[:, i].astype(int)]

                    all_stat_t2p.append(stat_t2p)
                    all_f_t2p.append(f_t2p)
                    all_ops.append(ops)
                    all_iscell_t2p.append(iscell_t2p)      
                    fneu_iscell_t2p.append(fneu_t2p)
                    spks_iscell_t2p.append(spks_t2p)  
                    if track_ops.nchannels==2:
                        fneu_chan2_iscell_t2p.append(fneu_chan2_t2p)
                        f_chan2_iscell_t2p.append(f_chan2_t2p)
                        redcell_iscell_t2p.append(redcell_t2p)  
        

        output_folderpath=os.path.join(folderpath, 'matched_suite2p')
        last_elements = [os.path.basename(path) for path in track_ops.all_ds_path]
        print(last_elements)
        
        # Save each element of each list to a .npy file
        for i in range(len(track_ops.all_ds_path)):
            stat_t2p, f_t2p, ops, iscell_t2p, fneu_t2p, spks_t2p = all_stat_t2p[i], all_f_t2p[i], all_ops[i], all_iscell_t2p[i], fneu_iscell_t2p[i], spks_iscell_t2p[i]
            subfolder_path = os.path.join(output_folderpath, last_elements[i])
            if not os.path.exists(subfolder_path):
                os.makedirs(subfolder_path)
            inner_folderpath = os.path.join(subfolder_path, 'suite2p')
            if not os.path.exists(inner_folderpath):
                os.makedirs(inner_folderpath)
            plane_folderpath = os.path.join(inner_folderpath, f'plane{j}')
            if not os.path.exists(plane_folderpath):
                os.makedirs(plane_folderpath)
            

            np.save(os.path.join(plane_folderpath,f"stat.npy"), stat_t2p)
            np.save(os.path.join(plane_folderpath,  f"F.npy"), f_t2p)
            np.save(os.path.join(plane_folderpath, f"ops.npy"), ops)
            np.save(os.path.join(plane_folderpath, f"iscell.npy"), iscell_t2p)
            np.save(os.path.join(plane_folderpath, f"Fneu.npy"), fneu_t2p)
            np.save(os.path.join(plane_folderpath,f"spks.npy"), spks_t2p)
            if track_ops.nchannels==2:
                for i, (redcell_t2p, f_chan2_t2p, fneu_chan2_t2p) in enumerate(zip(redcell_iscell_t2p, f_chan2_iscell_t2p, fneu_chan2_iscell_t2p)):
                    np.save(os.path.join(plane_folderpath,  f"F_chan2.npy"), f_chan2_t2p)
                    np.save(os.path.join(plane_folderpath,  f"Fneu_chan2.npy"), fneu_chan2_t2p)
                    np.save(os.path.join(plane_folderpath, f"redcell.npy"), redcell_t2p)


from track2p.gui.t2p_wd import Track2pWindow
from track2p.gui.import_wd import ImportWindow
from track2p.gui.raster_wd import RasterWindow
#from track2p.gui.s2p_wd import Suite2pWindow

class WindowManager:
    def __init__(self, main_window):
        self.main_window = main_window
        self.t2p_window = Track2pWindow(self.main_window)
        self.import_window = ImportWindow(self.main_window)
        self.raster_window=RasterWindow(self.main_window)
        #self.suite2p_window=Suite2pWindow(self.main_window)

    def open_track2p_wd(self):
        self.t2p_window.show()

    def open_import_wd(self):
        self.import_window.show()
        
    def open_raster_wd(self):
        self.raster_window.show()
        
        
        
    #def open_suite2p_wd(self):
    #   self.suite2p_window.show()



# contains code for generating plots based on track_ops object after the pipeline is run (this will be saved as an npy file)
# here all functions just take the track_ops object as input
import os

import numpy as np
import matplotlib.pyplot as plt

from skimage.exposure import match_histograms
from track2p.plot.utils import make_rgb_img, saturate_perc, get_all_wind_mean_img
from track2p.io.loaders import load_stat_ds_plane, get_all_roi_array_from_stat

def plot_reg_img_output(track_ops):
    # make a plot where on the top its all the images and the bottom is the overlays before and after registration
    nplanes = track_ops.nplanes
    n_row = nplanes + 2*nplanes # number of plays + 2 overlays per plane
    n_col = len(track_ops.all_ds_path) # number of datasets
    figsize = (10/3 * n_col, 10/3 * n_row)
    fig, axs = plt.subplots(n_row, n_col, figsize=figsize, dpi=300)

    # first populate first (n_planes) rows with images
    for i in range(nplanes):
        axs[i, 0].set_ylabel(f'plane{i}\nchan{track_ops.reg_chan}', rotation=0, size='large', ha='right', va='center', labelpad=20)
        for j in range(len(track_ops.all_ds_path)):
            img = track_ops.all_ds_avg_ch1[j][i] if track_ops.reg_chan==0 else track_ops.all_ds_avg_ch2[j][i]
            axs[i, j].imshow(img, cmap='gray', vmin=0, vmax=np.percentile(img, 99))


    for i in range(len(track_ops.all_ds_path)-1): # last one won't have overlay
        for j in range(nplanes):
            # get subplot indices
            row_nonreg = nplanes + 2*j # first shift for number of initial rows, then shift by 2 for each plane (before and after registration)
            row_reg = nplanes + 2*j + 1 # the row after the nonreg

            # get images
            ref_img = track_ops.all_ds_ref_img[i][j] # get ref image for this pair
            mov_img = track_ops.all_ds_mov_img[i][j] # get mov image for this pair
            mov_img_ref = track_ops.all_ds_mov_img_reg[i][j] # get mov image after registration for this pair

            # match histograms to reference
            mov_img = match_histograms(mov_img, ref_img)
            mov_img_ref = match_histograms(mov_img_ref, ref_img)

            # assemble and saturate the overlays
            img_rgb = make_rgb_img(ref_img, mov_img)
            img_rgb_reg = make_rgb_img(ref_img, mov_img_ref)
            img_rgb = saturate_perc(img_rgb, sat_perc=track_ops.sat_perc)
            img_rgb_reg = saturate_perc(img_rgb_reg, sat_perc=track_ops.sat_perc)

            # plot the overlays
            axs[row_nonreg, i].imshow(img_rgb)
            axs[row_reg, i].imshow(img_rgb_reg)

        if i == 0:
            for k in range(nplanes):
                axs[nplanes + 2*k,i].set_ylabel(f'plane{k}\nchan{track_ops.reg_chan}\nnon-reg', rotation=0, size='large', ha='right', va='center', labelpad=20)
                axs[nplanes + 2*k+1,i].set_ylabel(f'plane{k}\nchan{track_ops.reg_chan}\nreg', rotation=0, size='large', ha='right', va='center', labelpad=20)

            # loop through all subplots and remove ticks and labels and spines

    for ax in axs.flat:
        ax.set(xticks=[], yticks=[])
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['bottom'].set_visible(False)

    # add arrows and dashed lines etc.

    axs[2*nplanes, 0].annotate('', xy=(0, 1.1), xytext=(n_col + 0.5, 1.1), 
                xycoords='axes fraction', textcoords='axes fraction',
                arrowprops=dict(arrowstyle='-', linestyle='dashed', color='grey'), 
                annotation_clip=False)

    for i in range(len(track_ops.all_ds_path)-1):
        axs[nplanes-1,i].annotate('', xy=(0.5, -0.17), xytext=(0.5, -0.02), 
                        xycoords='axes fraction', textcoords='axes fraction',
                        arrowprops=dict(facecolor=(1,0,0), edgecolor=(1,0,0), shrink=0.05), 
                        annotation_clip=False)
        axs[nplanes-1,i+1].annotate('', xy=(-0.6, -0.17), xytext=(0.5, -0.02), 
                        xycoords='axes fraction', textcoords='axes fraction',
                        arrowprops=dict(facecolor=(0,1,0), edgecolor=(0,1,0), shrink=0.05), 
                        annotation_clip=False)
    

    # save figure into the output path
    fig.savefig(os.path.join(track_ops.save_path_fig, 'reg_img_output.png'), bbox_inches='tight', dpi=200)
    plt.close(fig)

def plot_roi_reg_output(track_ops):
    # make a plot where on the top its all the images and the bottom is the overlays before and after registration
    nplanes = track_ops.nplanes
    n_row = nplanes + 2*nplanes # number of plays + 2 overlays per plane
    n_col = len(track_ops.all_ds_path) # number of datasets
    figsize = (10/3 * n_col, 10/3 * n_row)
    # figsize = (10 * n_col, 10 * n_row)

    fig, axs = plt.subplots(n_row, n_col, figsize=figsize, dpi=300)

    # first populate first (n_planes) rows with images
    for i in range(nplanes):
        axs[i, 0].set_ylabel(f'plane{i}\nchan{track_ops.reg_chan}', rotation=0, size='large', ha='right', va='center', labelpad=20)
        for j in range(len(track_ops.all_ds_path)):
            img = track_ops.all_ds_avg_ch1[j][i]
            axs[i, j].imshow(img, cmap='gray', vmin=0, vmax=np.percentile(img, 99), alpha=0.5)
            # take the rois of reference unless it is last on, then take the rois of the mov
            all_roi = track_ops.all_ds_all_roi_array_ref[j][i] if j < len(track_ops.all_ds_path)-1 else track_ops.all_ds_all_roi_array_mov[j-1][i]
            for k in range(all_roi.shape[2]):
                axs[i, j].contour(all_roi[:,:,k], colors='C0', linewidths=0.3)

    # now populate the next (n_planes) rows with the overlays before registration
    for i in range(len(track_ops.all_ds_path)-1):
        print(f'Plotting contours for dataset {i}/{len(track_ops.all_ds_path)-1}')
        for j in range(track_ops.nplanes):

            row_nonreg = nplanes + 2*j # first shift for number of initial rows, then shift by 2 for each plane (before and after registration)
            row_reg = nplanes + 2*j + 1 # the row after the nonreg

            all_roi_array_ref = track_ops.all_ds_all_roi_array_ref[i][j]

            all_roi_array_mov = track_ops.all_ds_all_roi_array_mov[i][j]
            ref_mov_inters = track_ops.all_ds_ref_mov_inters[i][j]

            axs[row_nonreg,i].imshow(ref_mov_inters)
            for k in range(all_roi_array_ref.shape[2]):
                axs[row_nonreg,i].contour(all_roi_array_ref[:,:,k], colors='r', linewidths=0.3)
            for k in range(all_roi_array_mov.shape[2]):
                axs[row_nonreg,i].contour(all_roi_array_mov[:,:,k], colors='g', linewidths=0.3)

            # TODO: for reg row
            all_roi_array_reg = track_ops.all_ds_all_roi_array_reg[i][j]
            ref_reg_inters = track_ops.all_ds_ref_reg_inters[i][j]

            axs[row_reg,i].imshow(ref_reg_inters)
            for k in range(all_roi_array_ref.shape[2]):
                axs[row_reg,i].contour(all_roi_array_ref[:,:,k], colors='r', linewidths=0.3)
            for k in range(all_roi_array_reg.shape[2]):
                axs[row_reg,i].contour(all_roi_array_reg[:,:,k], colors='g', linewidths=0.3)

        if i == 0:
            for k in range(nplanes):
                axs[nplanes + 2*k,i].set_ylabel(f'plane{k}\nnon-reg', rotation=0, size='large', ha='right', va='center', labelpad=20)
                axs[nplanes + 2*k+1,i].set_ylabel(f'plane{k}\nreg', rotation=0, size='large', ha='right', va='center', labelpad=20)


    for ax in axs.flat:
        ax.set(xticks=[], yticks=[])
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['bottom'].set_visible(False)     

    # add arrows and dashed lines etc.

    axs[2*nplanes, 0].annotate('', xy=(0, 1.1), xytext=(n_col + 0.5, 1.1), 
                xycoords='axes fraction', textcoords='axes fraction',
                arrowprops=dict(arrowstyle='-', linestyle='dashed', color='grey'), 
                annotation_clip=False)

    for i in range(len(track_ops.all_ds_path)-1):
        axs[nplanes-1,i].annotate('', xy=(0.5, -0.17), xytext=(0.5, -0.02), 
                        xycoords='axes fraction', textcoords='axes fraction',
                        arrowprops=dict(facecolor='r', edgecolor='r', shrink=0.05), 
                        annotation_clip=False)
        axs[nplanes-1,i+1].annotate('', xy=(-0.6, -0.17), xytext=(0.5, -0.02), 
                        xycoords='axes fraction', textcoords='axes fraction',
                        arrowprops=dict(facecolor='g', edgecolor='g', shrink=0.05), 
                        annotation_clip=False)
    

    # save figure into the output path
    fig.savefig(track_ops.save_path_fig + 'reg_roi_output.png', bbox_inches='tight', dpi=200)     
    plt.close(fig)


def plot_thr_met_hist(all_ds_thr_met, all_ds_thr, track_ops):
    fig, axs = plt.subplots(track_ops.nplanes, len(all_ds_thr_met), figsize=(6*len(all_ds_thr_met), 6*track_ops.nplanes), sharey=True, sharex=True)
    for i in range(len(all_ds_thr_met)):
        for j in range(track_ops.nplanes):
            axs = np.array([axs]) if type(axs) is not np.ndarray else axs
            this_ax = axs[i] if track_ops.nplanes==1 else axs[j][i]
            n_reg_roi = len(all_ds_thr_met[i][j])
            n_abovethr_roi = np.sum(all_ds_thr_met[i][j]>all_ds_thr[i][j])
            this_ax.hist(all_ds_thr_met[i][j], bins=20)
            this_ax.axvline(all_ds_thr[i][j], color='grey', linestyle='--')
            # label the line with 'otsu threshold'
            this_ax.text(all_ds_thr[i][j]+0.02, this_ax.get_ylim()[1]*0.9, f'{track_ops.thr_method} thr.: {all_ds_thr[i][j]:.2f}')
            this_ax.set_title(f'ds{i} (ref) to ds{i+1} (reg); matched {n_abovethr_roi}/{n_reg_roi} ({n_abovethr_roi/n_reg_roi*100:.1f}%)')
            
            if i==0:
                this_ax.set_ylabel(f'ROI count (plane{j})')
    
    # remove the top and right spines
    for a in axs.flatten():
        a.spines['top'].set_visible(False)
        a.spines['right'].set_visible(False)

    
    plt.tight_layout()
    plt.savefig(os.path.join(track_ops.save_path_fig, 'thr_met_hist.png'), dpi=200)
    plt.close(fig)



def plot_n_matched_roi(all_ds_thr_met, all_ds_thr, track_ops):
    
    fig, axs = plt.subplots(track_ops.nplanes, 2, figsize=(8, 4*track_ops.nplanes), sharey='col', sharex=True)
                            
    for i in range(track_ops.nplanes):
        all_n_reg_roi = [len(all_ds_thr_met[j][i]) for j in range(len(all_ds_thr_met))]
        all_n_abovethr_roi = [np.sum(all_ds_thr_met[j][i]>all_ds_thr[j][i]) for j in range(len(all_ds_thr_met))]
        all_prop_abovethr = [n_abovethr_roi/n_reg_roi for n_abovethr_roi, n_reg_roi in zip(all_n_abovethr_roi, all_n_reg_roi)]

        ax0 = axs[i, 0] if track_ops.nplanes>1 else axs[0]
        
        ax0.bar(np.arange(len(all_n_reg_roi)), all_n_reg_roi, color='grey', label='total')
        ax0.bar(np.arange(len(all_n_abovethr_roi)), all_n_abovethr_roi, color='C0', label='above threshold')

        ax0.set_ylabel(f'ROI count (plane {i})')
        ax0.set_xticks(np.arange(len(all_n_reg_roi)))
        ax0.set_xticklabels([f'r{i}-{i+1}' for i in range(len(all_n_reg_roi))])
        ax0.legend(loc='lower right')
        ax0.spines['top'].set_visible(False)
        ax0.spines['right'].set_visible(False)

        ax1 = axs[i, 1] if track_ops.nplanes>1 else axs[1]

        ax1.plot(np.arange(len(all_n_reg_roi)), all_prop_abovethr, color='C0', marker='o')
        ax1.set_ylabel(f'ROI prop. (plane {i})')
        ax1.set_xticks(np.arange(len(all_n_reg_roi)))
        ax1.set_xticklabels([f'r{i}-{i+1}' for i in range(len(all_n_reg_roi))])
        ax1.set_ylim(0, 1)
        ax1.spines['top'].set_visible(False)
        ax1.spines['right'].set_visible(False)

        fig.savefig(os.path.join(track_ops.save_path_fig, 'n_prop_matched_roi.png'), dpi=200, bbox_inches='tight')
        plt.close(fig)

def plot_roi_match(all_ds_mean_img, all_ds_centroids, all_pl_match_mat, neuron_ids, track_ops, plane_idx=0, win_size=64, k=0, n=None, ch=1):

    if n is None:
        n = len(neuron_ids)

    nrows = len(neuron_ids)
    ncols = len(all_ds_mean_img)
    fig, axs = plt.subplots(nrows, ncols, figsize=(2*ncols, 2*nrows), dpi=50)

    for (i, nrn_id) in enumerate(neuron_ids):
        # check if neuron is not matched (if ith row of all_pl_match_mat is all None)
        if any(all_pl_match_mat[plane_idx][nrn_id,:]==None):
            continue
        
        # get all wind_mean_img (small window around centroid)
        all_wind_mean_img = get_all_wind_mean_img(all_ds_mean_img, all_ds_centroids, all_pl_match_mat, nrn_id, plane_idx=plane_idx, win_size=win_size)

        if i == 0:
            fig_ref = all_wind_mean_img[0] # reference image for matching histograms along whole image
        else:
            all_wind_mean_img[0] = match_histograms(all_wind_mean_img[0], fig_ref) # if not first ROI of whole image then match to first ROI of whole image

        for j in range(len(all_ds_mean_img)):
            wind_mean_img = all_wind_mean_img[j]
            ref_img = all_wind_mean_img[0]
            matched = match_histograms(wind_mean_img, ref_img)
            axs[i, j].imshow(matched, cmap='gray')
            # scatter middle pixel of the window
            axs[i, j].scatter(win_size/2, win_size/2, color='C0')
            if i==0:
                axs[i, j].set_title(f'ds {j} (pl {plane_idx})')
            if j==0:
                axs[i, j].set_ylabel(f'ROI {nrn_id} ({i+k}/{n})')

    # remove axes labels
    for ax in axs.flat:
        ax.set_xticks([])
        ax.set_yticks([])

    plt.tight_layout()
    if ch == 1:
        plt.savefig(os.path.join(track_ops.save_path_fig, f'roi_match_plane{plane_idx}_idx{k}-{k+len(neuron_ids)}.png'), dpi=50)
    elif ch == 2:
        plt.savefig(os.path.join(track_ops.save_path_fig, f'roi_match_chan2_plane{plane_idx}_idx{k}-{k+len(neuron_ids)}.png'), dpi=50)
    plt.close(fig)

def plot_roi_match_multiplane(all_ds_mean_img, all_ds_centroids, all_pl_match_mat, track_ops, win_size=48, ch=1):

    for i in range(track_ops.nplanes):
        pl_neuron_ids = np.arange(all_pl_match_mat[i].shape[0])
        for j in range(len(track_ops.save_path)):
            neuron_ids = pl_neuron_ids[~np.any(all_pl_match_mat[i]==None, axis=1)]

        if len(neuron_ids) < 100:
            plot_roi_match(all_ds_mean_img, all_ds_centroids, all_pl_match_mat, neuron_ids, track_ops, plane_idx=i, win_size=win_size, ch=ch)
        else:
            # plot in batches of 100
            for k in range(0, len(neuron_ids), 100):
                plot_roi_match(all_ds_mean_img, all_ds_centroids, all_pl_match_mat, neuron_ids[k:k+100], track_ops, plane_idx=i, win_size=win_size, k=k, n=len(neuron_ids), ch=ch)


def plot_allroi_match_multiplane(all_ds_mean_img, all_pl_match_mat, track_ops, ch=1):
    

    fig, axs = plt.subplots(track_ops.nplanes, len(track_ops.all_ds_path), figsize=(4*len(track_ops.all_ds_path), 4*track_ops.nplanes), dpi=300)

    for ds_idx, ds_path in enumerate(track_ops.all_ds_path):
        for plane_idx in range(track_ops.nplanes):
            # saturating the reference image
            ref_mean_img = all_ds_mean_img[0][plane_idx]
            perc = np.percentile(ref_mean_img, track_ops.sat_perc)
            ref_mean_img[ref_mean_img>perc] = perc

            ax = axs[ds_idx] if track_ops.nplanes==1 else axs[plane_idx, ds_idx]
            mean_img = all_ds_mean_img[ds_idx][plane_idx]
            mean_img_matched = match_histograms(mean_img, ref_mean_img)
            ax.imshow(mean_img_matched, cmap='gray')
            ax.set_title(f'ds{ds_idx} (plane{plane_idx})')

    # add contours
    for plane_idx in range(track_ops.nplanes):
        pl_neuron_ids = np.arange(all_pl_match_mat[plane_idx].shape[0])
        neuron_ids = pl_neuron_ids[~np.any(all_pl_match_mat[plane_idx]==None, axis=1)]
        neuron_colors = [np.random.rand(3) for i in range(len(neuron_ids))]

        for ds_idx, _ in enumerate(track_ops.all_ds_path):
            
            stat_ds_plane, _ = load_stat_ds_plane(track_ops.all_ds_path[ds_idx], track_ops, plane_idx=plane_idx)
            all_roi_array = get_all_roi_array_from_stat(stat_ds_plane, track_ops)
            
            ax = axs[ds_idx] if track_ops.nplanes==1 else axs[plane_idx, ds_idx]
        
            for (i, neuron_idx) in enumerate(neuron_ids):
                idx_orig = all_pl_match_mat[plane_idx][neuron_idx, ds_idx] # neurons index in the original recording
                cont = stat_ds_plane

                cont_plot = ax.contour(all_roi_array[:,:,idx_orig], linewidths=0.5)

                for collection in cont_plot.collections:
                    collection.set_edgecolor(neuron_colors[i])  # RGB value for red
        
        left_axs = axs[0] if track_ops.nplanes==1 else axs[plane_idx,0]
        left_axs.set_ylabel(f'plane{plane_idx} (n={len(neuron_ids)})')

    # remove axes elements
    for ax in axs.flat:
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xticklabels([])
        ax.set_yticklabels([])

    plt.tight_layout()
    if ch == 1:
        plt.savefig(os.path.join(track_ops.save_path_fig, f'all_roi_match.png'), dpi=300)
    else:
        plt.savefig(os.path.join(track_ops.save_path_fig, f'all_roi_match_chan2.png'), dpi=300)
    plt.close(fig)



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QTabWidget, QVBoxLayout, QWidget, QSplitter, QHBoxLayout, QFrame, QFrame
from track2p.gui.fluo_plot import FluorescencePlotWidget
from track2p.gui.roi_plot import ZoomPlotWidget
from track2p.gui.cell_plot import CellPlotWidget
from track2p.gui.data_management import DataManagement
from track2p.gui.raster_wd import RasterWindow

class CentralWidget(QWidget):
    def __init__(self, main_window):
        super().__init__()

        self.main_window = main_window
        self.fluorescences_plotting = None
        self.rois_plotting = None
        self.selected_roi = None
        self.cell_plot = None
        self.track_ops_dict=None
        self.data_management = DataManagement(self)
        self.vector_curation_t2p = self.data_management.vector_curation_t2p
        self.init_central_widget()

    def init_central_widget(self):
        self.top = QFrame()
        self.top.setFrameShape(QFrame.StyledPanel)
        self.top_layout = QHBoxLayout(self.top)

        self.top_right = QFrame()
        self.top_right.setFrameShape(QFrame.StyledPanel)
        self.top_layout_right = QVBoxLayout(self.top_right)

        self.tabs = QTabWidget(self)

        self.splitter1 = QSplitter(Qt.Horizontal)
        self.splitter1.addWidget(self.tabs)
        self.splitter1.addWidget(self.top)
        self.splitter1.setSizes([100, 100])

        self.splitter2 = QSplitter(Qt.Horizontal)
        self.splitter2.addWidget(self.top_right)

        self.splitter3 = QSplitter(Qt.Vertical)
        self.splitter3.addWidget(self.splitter1)
        self.splitter3.addWidget(self.splitter2)
        self.splitter3.setSizes([100, 100])

        central_layout = QVBoxLayout()
        central_layout.addWidget(self.splitter3)
        self.setLayout(central_layout)
        
        
    def create_mean_img(self,channel):
        for i, (ops, stat_t2p) in enumerate(zip(self.data_management.all_ops, self.data_management.all_stat_t2p)):
            tab = QWidget()
            self.cell_plot = CellPlotWidget(tab, ops=ops, stat_t2p=stat_t2p, f_t2p=self.data_management.all_f_t2p[i],
                                       colors=self.data_management.colors, update_selection_callback=self.update_selection,
                                       all_f_t2p=self.data_management.all_f_t2p, all_ops=self.data_management.all_ops, channel=channel)
            layout = QVBoxLayout(tab)
            layout.addWidget(self.cell_plot)
            tab.setLayout(layout)
            self.tabs.addTab(tab, f"Day {i + 1}")
            self.cell_plot.cell_selected.connect(self.update_selection)
            
    
    def create_mean_img_from_curation(self):
        import_window = self.main_window.window_manager.import_window
        t2p_window = self.main_window.window_manager.t2p_window

        if import_window is not None and import_window.plane is not None:
            self.data_management.import_files(import_window.path_to_t2p, import_window.plane)
        elif t2p_window is not None and t2p_window.saved_directory is not None:
            self.data_management.import_files(t2p_window.saved_directory, t2p_window.plane)
        else:
            print("Both import_window and t2p_window are None or not properly initialized.")
        
    def clear(self):
        self.data_management.reset_attributes()
        if self.fluorescences_plotting:
            self.top_layout_right.removeWidget(self.fluorescences_plotting)
            self.fluorescences_plotting.deleteLater()
            self.fluorescences_plotting = None
        if self.rois_plotting:
            self.top_layout.removeWidget(self.rois_plotting)
            self.rois_plotting.deleteLater()
            self.rois_plotting = None
        for i in range(self.tabs.count()):
            self.tabs.removeTab(0)
            
            
    def update_selection(self, selected_cell_index):
        self.selected_roi = selected_cell_index
        self.main_window.status_bar.spin_box.setValue(selected_cell_index) 
        self.main_window.status_bar.roi_state_value.setText(f"{self.vector_curation_t2p[selected_cell_index]}")        
        #it removes the underline of the previsouly selected cell even if the tab is not visible (not the current tab)
        for i in range(self.tabs.count()): 
            tab_widget = self.tabs.widget(i)
            cell_object = tab_widget.findChild(CellPlotWidget)
            cell_object.remove_previous_underline()
        current_tab_index = self.tabs.currentIndex()
        current_tab_widget = self.tabs.widget(current_tab_index)
        cell_plot = current_tab_widget.findChild(CellPlotWidget)
        if cell_plot:
            cell_plot.underline_cell(selected_cell_index)
        if self.fluorescences_plotting is None:
            self.fluorescences_plotting = FluorescencePlotWidget(all_f_t2p=self.data_management.all_f_t2p,
                                                           all_ops=self.data_management.all_ops,
                                                          colors=self.data_management.colors)
            self.top_layout_right.addWidget(self.fluorescences_plotting)
        if self.rois_plotting is None:
            self.rois_plotting = ZoomPlotWidget(all_ops=self.data_management.all_ops,
                                            all_stat_t2p=self.data_management.all_stat_t2p,
                                            colors=self.data_management.colors,
                                            all_iscell_t2p=self.data_management.all_iscell,
                                            t2p_match_mat_allday=self.data_management.t2p_match_mat_allday,track_ops=self.track_ops)
            self.top_layout.addWidget(self.rois_plotting)
        
        self.fluorescences_plotting.display_all_f_t2p(selected_cell_index)
        self.rois_plotting.display_zooms(selected_cell_index)
            
                   
    def display_first_ROI(self,index):
        """it displays the first cell of the t2p_match_mat_allday and its fluorescence and zooms across days. It is called when the application is opened.
        An instance of FluorescencePlotWidget and an instance of ZoomPlotWidget are created and added to attributes of the MainWindow class. """
        tab_widget = self.tabs.widget(0)
        cell_object = tab_widget.findChild(CellPlotWidget) #It finds the instance of the CellPlotWidget class in the first tab of the QTabWidget
        cell_object.underline_cell(index)
        cell_object.draw()
        if self.fluorescences_plotting is None:
            self.fluorescences_plotting = FluorescencePlotWidget(all_f_t2p=self.data_management.all_f_t2p,
                                                           all_ops=self.data_management.all_ops,
                                                           colors=self.data_management.colors, all_stat_t2p=self.data_management.all_stat_t2p)
            self.top_layout_right.addWidget(self.fluorescences_plotting)
        if self.rois_plotting is None:
            self.rois_plotting = ZoomPlotWidget(all_ops=self.data_management.all_ops,
                                            all_stat_t2p=self.data_management.all_stat_t2p,
                                            colors=self.data_management.colors,
                                            all_iscell_t2p=self.data_management.all_iscell,
                                            t2p_match_mat_allday=self.data_management.t2p_match_mat_allday,track_ops=self.data_management.track_ops, imgs= self.cell_plot.all_img)
            self.top_layout.addWidget(self.rois_plotting)
        self.fluorescences_plotting.display_all_f_t2p(index)
        self.rois_plotting.display_zooms(index)
        self.main_window.status_bar.roi_state_value.setText(f"{self.vector_curation_t2p[self.main_window.status_bar.spin_box.value()]}") #



from PyQt5.QtWidgets import QToolBar,QMenu,QAction,QToolButton

class Toolbar(QToolBar):
    def __init__(self, main_window):
        super().__init__()

        self.main_window = main_window
        self.init_tool_bar()

    def init_tool_bar(self):
        data_menu = QMenu(self)
        run_menu = QMenu(self)
        visualization_menu = QMenu(self)

        import_action = QAction("Load processed data (⌘L or Ctrl+L)", self)
        import_action.setShortcut("Ctrl+L")
        import_action.triggered.connect(self.main_window.window_manager.open_import_wd)
        data_menu.addAction(import_action)

        track2p_action = QAction("Run track2p algorithm (⌘R or Ctrl+R)", self)
        track2p_action.setShortcut("Ctrl+R")
        track2p_action.triggered.connect(self.main_window.window_manager.open_track2p_wd)
        run_menu.addAction(track2p_action)

        raster_action = QAction("Generate raster plot (⌘G or Ctrl+G)", self)
        raster_action.setShortcut("Ctrl+G")
        raster_action.triggered.connect(self.main_window.window_manager.open_raster_wd)
        visualization_menu.addAction(raster_action)

        self.add_tool_menu("File", data_menu)
        self.add_tool_menu("Run", run_menu)
        self.add_tool_menu("Visualization", visualization_menu)

    def add_tool_menu(self, name, menu):
        button = QToolButton(self)
        button.setText(name)
        button.setMenu(menu)
        button.setPopupMode(QToolButton.InstantPopup)
        button.setStyleSheet("font-size: 13px")
        self.addWidget(button)
        
    


from PyQt5.QtWidgets import QApplication
import time 
from track2p.gui.main_wd import MainWindow
from PyQt5.QtGui import QIcon
import os

# the same script as track2p/gui/run_gui.py

if __name__ == '__main__':
    start_time = time.time()

    app = QApplication([])
 

    # Utiliser un chemin relatif pour l'icône
    icon_path = os.path.join(os.path.dirname(__file__), 'resources', 'logo.png')
    print(icon_path)
    
    if not os.path.exists(icon_path):
        print(f"Icon file not found: {icon_path}")
    else:
        app.setWindowIcon(QIcon(icon_path))
    
    mainWindow = MainWindow()
    mainWindow.setWindowTitle("track2p")
   
    end_time = time.time()
    print(f"Application took {end_time - start_time} seconds to open.")
    app.exec_()


# contains code for generating plots while the pipeline is running

import numpy as np
import matplotlib.pyplot as plt

from track2p.plot.utils import match_hist_all

def plot_all_planes(all_ds_avg_ch, track_ops, sat_perc=99, ch='funcional'):
    nplanes = track_ops.nplanes
    fig, axs = plt.subplots(nplanes, len(track_ops.all_ds_path), figsize=(3 * len(track_ops.all_ds_path), 3 * nplanes), dpi=300)
    # add dummy dimension to axs if only one plane
    if nplanes==1:
        axs = np.expand_dims(axs, axis=0)
    all_ds_avg_ch_matched = match_hist_all(all_ds_avg_ch)


    for i in range(nplanes):
        for j in range(len(track_ops.all_ds_path)):
            img = all_ds_avg_ch_matched[j][i]
            axs[i, j].imshow(img, cmap='gray', vmin=0, vmax=np.percentile(img, sat_perc))
            axs[i, j].set_title('Plane ' + str(i) + ' in dataset ' + str(j))
            axs[i, j].axis('off')

    fig.savefig(track_ops.save_path_fig + f'mean_fov_{ch}.png', bbox_inches='tight', dpi=200)     

    plt.close(fig)

    




import itk
import numpy as np

def reg_img_elastix(ref_img, mov_img, track_ops):
    # convert to itk images
    ref_img_itk = itk.GetImageFromArray(ref_img)
    mov_img_itk = itk.GetImageFromArray(mov_img)

    # import default parameter map
    parameter_object = itk.ParameterObject.New()
    parameter_map = parameter_object.GetDefaultParameterMap(track_ops.transform_type)
    parameter_object.AddParameterMap(parameter_map)

    # call registration function
    mov_img_reg_itk, reg_params = itk.elastix_registration_method(
        ref_img_itk,
        mov_img_itk,
        parameter_object = parameter_object
    )

    # convert back to numpy array
    mov_img_reg = itk.GetArrayFromImage(mov_img_reg_itk)
    reg_params.SetParameter("FinalBSplineInterpolationOrder", "0")

    return mov_img_reg, reg_params

def itk_reg_roi(roi, reg_params):
    roi_itk = itk.GetImageFromArray(roi.astype(np.uint8))
    roi_itk_trans = itk.transformix_filter(roi_itk, reg_params)
    roi_trans = itk.GetArrayFromImage(roi_itk_trans)
    return roi_trans

def itk_reg_all_roi(all_roi, reg_params):
    all_roi_array_reg = np.zeros_like(all_roi)
    for i in range(all_roi.shape[2]):
        roi_array = all_roi[:,:,i]
        roi_array_reg = itk_reg_roi(roi_array, reg_params)
        all_roi_array_reg[:,:,i] = roi_array_reg
    return all_roi_array_reg


import time 
from qtpy.QtCore import Signal
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.pyplot as plt
import numpy as np
import skimage

class CellPlotWidget(FigureCanvas):
    '''This class is used to view and interact with the mean image of each recording (day)'''
    cell_selected = Signal(int)

    def __init__(self, tab=None, ops=None, stat_t2p=None, f_t2p=None, colors=None, update_selection_callback=None,
                 all_f_t2p=None, all_stat_t2p=None, all_ops=None, initial_colors=None, channel=None):
        """It initializes the class attributes and connects certain events to their respective handlers. It also creates the figure and the axes to display the mean image of the recording."""
        self.fig, self.ax_image = plt.subplots(1, 1) 
        self.fig.set_facecolor('black')
        super().__init__(self.fig)
        self.ops = ops
        self.stat_t2p = stat_t2p
        self.f_t2p = f_t2p
        self.all_fluorescence = all_f_t2p
        self.all_stat_t2p=all_stat_t2p
        self.all_ops=all_ops
        self.colors = colors
        self.initial_colors=initial_colors
        self.channel = channel
        self.all_img, self.img= self.load_all_imgs()
        self.selected_cell_index = None
        self.mpl_connect('button_press_event', self.on_mouse_press)
        self.update_selection_callback = update_selection_callback
        self.nb_cells= len(self.stat_t2p)
        self.plot_cells()
        self.initialize_interactions()

    def load_all_imgs(self):
        print('loading all images')
        print('channel img:', self.channel)
        all_img = []
        img= None 
        if self.channel == 'max_proj':
            print('max_proj')
            img= self.ops['max_proj']
            Ly = self.ops['Ly'] # number of pixels in y (512)
            Lx = self.ops['Lx'] # number of pixels in x (512)
            yr = self.ops['yrange'] # first and last pixel of crop in y
            xr = self.ops['xrange'] # first and last pixel of crop in x
            range_img = np.zeros((Ly, Lx))
            range_img[yr[0]:yr[1], xr[0]:xr[1]] = img
            img= range_img
            for ops in self.all_ops:
                img_ops= ops['max_proj']
                Ly = ops['Ly']
                Lx = ops['Lx']
                yr = ops['yrange']
                xr = ops['xrange']
                range_img = np.zeros((Ly, Lx))
                range_img[yr[0]:yr[1], xr[0]:xr[1]] = img_ops
                img_ops= range_img
                all_img.append(img_ops)
        if self.channel == 'Vcorr':
            print('Vcorr')
            img= self.ops['max_proj']
            Ly = self.ops['Ly'] # number of pixels in y (512)
            Lx = self.ops['Lx'] # number of pixels in x (512)
            yr = self.ops['yrange'] # first and last pixel of crop in y
            xr = self.ops['xrange'] # first and last pixel of crop in x
            range_img = np.zeros((Ly, Lx))
            range_img[yr[0]:yr[1], xr[0]:xr[1]] = img
            img= range_img
            for ops in self.all_ops:
                img_ops= ops['max_proj']
                Ly = ops['Ly']
                Lx = ops['Lx']
                yr = ops['yrange']
                xr = ops['xrange']
                range_img = np.zeros((Ly, Lx))
                range_img[yr[0]:yr[1], xr[0]:xr[1]] = img_ops
                img_ops= range_img
                all_img.append(img_ops)            
        if self.channel == '0':
            #print('0')
            all_img = [ops['meanImg'] for ops in self.all_ops]
            img= self.ops['meanImg']
        if self.channel == '1':
            #print('1')
            all_img = [ops['meanImg_chan2'] for ops in self.all_ops]
            img= self.ops['meanImg_chan2']

        return all_img, img
        
    def plot_cells(self):
        """It plots the mean image of the recording and the contours of the cells. It also sets the axis to be invisible and the title of the plot. It uses the colors attribute to color the contours of the cells. 
        the match_histograms function of the skimage library is used to match the histograms of the mean image of the recording and the last mean image of the recordings. . This is done to make the mean images of the recordings more comparable."""
        self.ax_image.clear()
        start = time.time()
        match_mean_img=skimage.exposure.match_histograms(self.img,self.all_img[-1], channel_axis=None)
        self.ax_image.imshow(match_mean_img, cmap='gray')
        cell_count = 0
        for cell in range(self.nb_cells):
           # print(cell)
            bin_mask = np.zeros_like(self.img) #create a binary mask with the same shape as the mean image of the recording
            bin_mask[self.stat_t2p[cell]['ypix'], self.stat_t2p[cell]['xpix']] = 1
            color_cell=self.colors[cell]
            self.ax_image.contour(bin_mask, levels=[0.5], colors=[color_cell], linewidths=1) 
            cell_count += 1
        self.ax_image.axis('off')
        print(f'time for plotting cells on mean image for recording : {time.time()-start}')
        print(f'Total cells plotted: {cell_count}')
        self.draw()
        
    def plot_cells_remix(self,keys):
        self.ax_image.clear()
        start = time.time()
        match_mean_img=skimage.exposure.match_histograms(self.img,self.all_img[-1], channel_axis=None)
        self.ax_image.imshow(match_mean_img, cmap='gray')
        cell_count = 0
        for cell in range(self.nb_cells):
            if cell in keys:
                continue
            bin_mask = np.zeros_like(self.img) #create a binary mask with the same shape as the mean image of the recording
            bin_mask[self.stat_t2p[cell]['ypix'], self.stat_t2p[cell]['xpix']] = 1
            color_cell=self.colors[cell]
            self.ax_image.contour(bin_mask, levels=[0.5], colors=[color_cell], linewidths=1) 
            cell_count += 1
        self.ax_image.axis('off')
        print(f'time for plotting cells on mean image for recording : {time.time()-start}')
        print(f'Total cells plotted: {cell_count}')
        self.draw()
        
        
           
    def underline_cell_remix(self,colors):
        
        for cell in range(self.nb_cells):
            bin_mask = np.zeros_like(self.img)
            bin_mask[self.stat_t2p[cell]['ypix'], self.stat_t2p[cell]['xpix']] = 1
            color_cell=colors[cell]
            self.ax_image.contour(bin_mask, levels=[0.5], colors=[color_cell], linewidths=1)
        self.draw()

   
        
    def underline_cell(self,selected_cell_index):
        """It underlines the selected cell by increasing the linewidth of the contour of the cell"""
        for cell in range(self.nb_cells):
            if cell == selected_cell_index:
                bin_mask = np.zeros_like(self.img) 
                bin_mask[self.stat_t2p[cell]['ypix'], self.stat_t2p[cell]['xpix']] = 1 
                color_cell=self.colors[cell]
                self.ax_image.contour(bin_mask, levels=[0.5], colors=[color_cell], linewidths=3)
        self.draw() 
    
    def remove_previous_underline(self):
        """It removes the underline of the previously selected cell by decreasing the linewidth of the contour of the cell."""
        for collection in self.ax_image.collections:
            collection.set_linewidth(1)
         # important, don't forget it ! (update the plot)
            
    def initialize_interactions(self):
        """This method is used to initialize user interactions with the mean image. It connects the scroll event to the on_scroll method and records the initial xlim and ylim of the mean image."""
        self.cid_scroll = self.fig.canvas.mpl_connect('scroll_event', self.on_scroll)
        self.initial_xlim = self.ax_image.get_xlim()
        self.initial_ylim = self.ax_image.get_ylim()

    def on_scroll(self, event):
        """it allows to zoom in and out of the mean image of the recording. It is called when the mouse wheel is scrolled. It uses the base_scale to zoom in and out of the mean image of the recording. """
        if event.inaxes == self.ax_image:
            current_xlim = self.ax_image.get_xlim()
            current_ylim = self.ax_image.get_ylim()
            base_scale = 0.9
            scale_factor = base_scale if event.button == 'up' else 1/base_scale
            # Get the coordinates of the mouse cursor in data coordinates
            x_data, y_data = event.xdata, event.ydata
            # Compute the new limits centered on the mouse cursor
            new_xlim = [x_data - (x_data - x) * scale_factor for x in current_xlim]
            new_ylim = [y_data - (y_data - y) * scale_factor for y in current_ylim]            
            # Ensure the zoom doesn't exceed the initial image boundaries
            new_xlim = [max(self.initial_xlim[0], min(self.initial_xlim[1], x)) for x in new_xlim]
            new_ylim = [max(self.initial_ylim[1], min(self.initial_ylim[0], y)) for y in new_ylim]

            self.ax_image.set_xlim(new_xlim)
            self.ax_image.set_ylim(new_ylim)
            self.fig.canvas.draw_idle()

    def on_mouse_press(self, event):
        """It allows to select a cell by clicking on it. It is called when the mouse is clicked. It uses the x and y coordinates of the mouse cursor to determine if a cell is clicked. If a cell is clicked, the selected_cell_index attribute is updated and the cell_selected signal is emitted.
        The update_selection method of the MainWindow class is called and used to update the fluorescence and zoom plots with the selected cell. """
        start = time.time()
        if event.inaxes == self.ax_image:
            x, y = event.xdata, event.ydata
            for j, cell_info in enumerate(self.stat_t2p):
                ypix = cell_info['ypix'] #ypix are the y coordinates of the pixels of the cell
                xpix = cell_info['xpix'] #xpix are the x coordinates of the pixels of the cell
                if np.any((xpix == int(x)) & (ypix == int(y))):
                    self.selected_cell_index = j
                    self.update_selection_callback(j)
                    print(f"Cell selected: {j}", flush=True)
                    break
        print(f'time taken for updating: {time.time()-start}')
        
        



import os
import io
from PyQt5.QtWidgets import QVBoxLayout, QWidget, QPushButton, QFileDialog, QLineEdit, QLabel, QFormLayout, QCheckBox, QComboBox,QGraphicsView,QGraphicsScene,QSplitter,QGroupBox
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPixmap
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from scipy.stats import zscore
from sklearn.decomposition import PCA
from openTSNE import TSNE
import matplotlib.pyplot as plt
import copy



class RasterWindow(QWidget):
        #QWidget is the parent class
        def __init__(self, mainWindow ):
            super(RasterWindow,self).__init__()
            self.main_window = mainWindow
            self.raster_type=None
            self.bin_size = None
            self.filename=None
            self.all_f_t2p_preproc=None
            self.vmin_value=None
            self.vmax_value=None
   
            
            
            #Create the right-hand side of the window
            layout = QFormLayout()
            
            #label_imp= QLabel("Import the directory containing the 'track2p' folder:")
            #field_imp=QPushButton("Import")
            #field_imp.clicked.connect(self.load_directory_contents)
            #layout.addRow(label_imp,field_imp)
            
            #label_imp_path= QLabel("Here is the path of the imported directory:")
            #self.field_imp_path=QLabel()
           # layout.addRow(label_imp_path,self.field_imp_path)
            
            #label_plane= QLabel("Choose the plane to analyze:")
            #self.field_plane=QLineEdit()
            #self.field_plane.setText('0')
            #self.field_plane.setFixedWidth(50)
            #layout.addRow(label_plane,self.field_plane)
            
            self.checkboxes=[]
            label_checkbox= QLabel("Choose the sorting method:")
            field_checkbox= QVBoxLayout()
            self.checkbox1 = QCheckBox('without sorting', self)
            self.checkbox2 = QCheckBox('sorting by PCA', self)
            self.checkbox3 = QCheckBox('sorting by PCA on given day', self)
            self.checkbox4 = QCheckBox('sorting by tSNE', self)
            self.checkbox5 = QCheckBox('sorting by tSNE on given day', self)
            self.checkboxes.append(self.checkbox1)
            self.checkboxes.append(self.checkbox2)
            self.checkboxes.append(self.checkbox3)
            self.checkboxes.append(self.checkbox4)
            self.checkboxes.append(self.checkbox5)

            self.checkbox3.stateChanged.connect(self.update_day_choice)
            self.checkbox5.stateChanged.connect(self.update_day_choice)

            self.day_choice= QComboBox(self)
            self.day_choice.addItem('Choose recording index (for sorting on given day)')
            field_checkbox.addWidget(self.checkbox1)
            field_checkbox.addWidget(self.checkbox2)
            field_checkbox.addWidget(self.checkbox3)
            field_checkbox.addWidget(self.checkbox4)
            field_checkbox.addWidget(self.checkbox5)
            field_checkbox.addWidget(self.day_choice)
            layout.addRow(label_checkbox,field_checkbox)

            for checkbox in self.checkboxes:
                checkbox.stateChanged.connect(self.handle_checkbox_state)
      
            self.combined_checkbox=QCheckBox('Combined (put neurons of all planes together) ', self)
            self.combined_checkbox.stateChanged.connect(self.concatenate_rasters)
            layout.addRow(" ",self.combined_checkbox)

            label_check=QLabel("Advanced options:")
            self.check=QCheckBox(self)
            self.check.stateChanged.connect(self.display_advanced_options)
            
            layout.addRow(label_check,self.check)
            self.advanced_options_group = QGroupBox()
            self.advanced_options_group.setVisible(False)
            group_layout = QFormLayout()
            self.bin= QLineEdit()
            self.bin.setText('1')
            self.bin.setFixedWidth(50)
            group_layout.addRow("Averaging bin size:", self.bin)
            self.vmin= QLineEdit()
            self.vmin.setText('0')
            self.vmin.setFixedWidth(50)
            group_layout.addRow("vmin:", self.vmin)
            self.vmax= QLineEdit()
            self.vmax.setText('1.96')
            self.vmax.setFixedWidth(50)
            group_layout.addRow("vmax:", self.vmax)
            self.advanced_options_group.setLayout(group_layout)
            layout.addRow('',self.advanced_options_group)
            

            # Initially hide the widgets
            self.bin.setVisible(False)
            self.vmin.setVisible(False)
            self.vmax.setVisible(False)         
            
            label_run= QLabel("Run the analysis:")
            field_run=QPushButton("Run")
            field_run.clicked.connect(self.run)
            layout.addRow(label_run,field_run)
            
            label_save= QLabel("Save the figure:")
            field_save = QPushButton("Save")
            field_save.clicked.connect(self.get_output_save_path)
            layout.addRow(label_save,field_save)
            
            label_output_path= QLabel("The figure has been saved here:")
            self.field_output_path=QLabel()
            layout.addRow(label_output_path,self.field_output_path)
            
            self.view=QGraphicsView(self) #QGraphicsView is a subclass of QWidget
            
            splitter = QSplitter(Qt.Horizontal)            
            right_widget = QWidget()
            right_widget.setLayout(layout)
            splitter.addWidget(right_widget)
            splitter.addWidget(self.view)
            
            main_layout = QVBoxLayout(self)
            main_layout.addWidget(splitter)
            
            self.setLayout(main_layout)
          
############################################################################################################################################################################

        def update_day_choice(self):
            if self.day_choice.count() ==1:
                self.all_stat_t2p=self.main_window.central_widget.data_management.all_stat_t2p
                if self.checkbox3.isChecked() or self.checkbox5.isChecked():
                    print(self.day_choice.count())
                    self.day_choice.clear()
                    for i in range(len(self.all_stat_t2p)):
                        self.day_choice.addItem(str(i + 1))
                print('ComboBox updated')

        def handle_checkbox_state(self):
            sender = self.sender()
            if sender.isChecked():
                for checkbox in self.checkboxes:
                    if checkbox != sender:
                        checkbox.setChecked(False)

        def run(self):
            if self.combined_checkbox.isChecked():
                self.all_f_t2p= self.concatenate_rasters()
            else: 
                self.all_f_t2p=self.main_window.central_widget.data_management.all_f_t2p
            self.plane=self.main_window.central_widget.data_management.plane
            self.preprocessing()
            self.get_checkbox_choice()


        def display_advanced_options(self,state):
            if state == Qt.Checked:
                self.advanced_options_group.setVisible(True)
                self.bin.setVisible(True)
                self.vmin.setVisible(True)
                self.vmax.setVisible(True)
            else:
                self.advanced_options_group.setVisible(False)
                self.bin.setVisible(False)
                self.vmin.setVisible(False)
                self.vmax.setVisible(False)

        def fit_pca_1d(self,data):
            print('fitting 1d-PCA...')
            pca=PCA(n_components=1)
            pca.fit(data)
            embedding = pca.components_.T
            return embedding

        def fit_tsne_1d(self,data):
            print('fitting 1d-tSNE...')
            tsne = TSNE(
            n_components=1,
            perplexity=30,
            initialization="pca",
            metric="euclidean",
            n_jobs=8,
            random_state=3
            )

            tsne_emb = tsne.fit(data.T)
            return tsne_emb

        
        def get_output_save_path(self):
            raster_folder = os.path.join(self.load_directory,'track2p', 'raster')
            os.makedirs(raster_folder, exist_ok=True)
            self.field_output_path.setText(f'{raster_folder}')
            filename=os.path.join(raster_folder,'rasterplot_{}_plane{}.pdf'.format(self.raster_type,str(self.field_plane.text())))
            plt.savefig(filename)
                
        def get_checkbox_choice(self):
            vmin=float(self.vmin.text())
            vmax=float(self.vmax.text())
            if self.checkbox1.isChecked():
                self.raster_type='without_sorting'
                self.plot_track2p_rasters(self.all_f_t2p_preproc, bin_size=self.bin_size, vmin=vmin, vmax=vmax)
            if self.checkbox2.isChecked():
                self.raster_type='sorting_by_PCA'
                all_pca_emb_1d = []
                for f in tqdm(self.all_f_t2p_preproc):
                    print(f.shape)
                    pca_emb_1d = self.fit_pca_1d(f.T)
                    print(pca_emb_1d.shape)
                    all_pca_emb_1d.append(pca_emb_1d)
                all_f_t2p_sorted = []
                for (i, f) in enumerate(self.all_f_t2p_preproc):
                    sort_inds = np.argsort(np.array(all_pca_emb_1d[i]).squeeze())
                    f_sorted = f[sort_inds, :].squeeze()
                    all_f_t2p_sorted.append(f_sorted)
                self.plot_track2p_rasters(all_f_t2p_sorted, bin_size=self.bin_size, vmin=vmin, vmax=vmax) # plot the rasters
            if self.checkbox3.isChecked():
                self.raster_type='sorting_by_PCA_and_by_day_' + str(self.day_choice.currentText())
                all_pca_emb_1d = []
                for f in tqdm(self.all_f_t2p_preproc):
                    print(f.shape)
                    pca_emb_1d = self.fit_pca_1d(f.T)
                    print(pca_emb_1d.shape)
                    all_pca_emb_1d.append(pca_emb_1d)
                all_f_t2p_sorted = []
                for (i, f) in enumerate(self.all_f_t2p_preproc):
                    user_i= int(self.day_choice.currentText()) - 1
                    sort_inds = np.argsort(np.array(all_pca_emb_1d[user_i]).squeeze())
                    f_sorted = f[sort_inds, :].squeeze()
                    all_f_t2p_sorted.append(f_sorted)
                self.plot_track2p_rasters(all_f_t2p_sorted, bin_size=self.bin_size, vmin=vmin, vmax=vmax)
            if self.checkbox4.isChecked():
                self.raster_type='sorting_by_tSNE'
                all_tsne_emb_1d = []
                for f in tqdm(self.all_f_t2p_preproc):
                    print(f.shape)
                    tsne_emb_1d = self.fit_tsne_1d(f.T)
                    all_tsne_emb_1d.append(tsne_emb_1d)
                all_f_t2p_sorted = []
                for (i, f) in enumerate(self.all_f_t2p_preproc):
                    sort_inds = np.argsort(np.array(all_tsne_emb_1d[i]).squeeze())
                    f_sorted = f[sort_inds, :].squeeze()
                    all_f_t2p_sorted.append(f_sorted)
                self.plot_track2p_rasters(all_f_t2p_sorted, bin_size=self.bin_size, vmin=vmin, vmax=vmax) # plot the rasters
            if self.checkbox5.isChecked():
                self.raster_type='sorting_tSNE_and_by_day_' + str(self.day_choice.currentText())
                all_tsne_emb_1d = []
                for f in tqdm(self.all_f_t2p_preproc):
                    print(f.shape)
                    tsne_emb_1d = self.fit_tsne_1d(f.T)
                    all_tsne_emb_1d.append(tsne_emb_1d)
                all_f_t2p_sorted = []
                for (i, f) in enumerate(self.all_f_t2p_preproc):
                    user_i= int(self.day_choice.currentText()) - 1
                    sort_inds = np.argsort(np.array(all_tsne_emb_1d[user_i]).squeeze())
                    f_sorted = f[sort_inds, :].squeeze()
                    all_f_t2p_sorted.append(f_sorted)
                self.plot_track2p_rasters(all_f_t2p_sorted, bin_size=self.bin_size, vmin=vmin, vmax=vmax) # plot the rasters

       
        def preprocessing(self):
            bin_data = True
            self.bin_size = int(self.bin.text()) #number of frames to average (1 = no averging)
            print(f'bin_size: {self.bin_size}')
            rem_zero_rows = True
            if bin_data:
                all_f_t2p_original = copy.deepcopy(self.all_f_t2p)
                self.all_f_t2p_preproc = [np.mean(f.reshape(f.shape[0], -1, self.bin_size), axis=2) for f in all_f_t2p_original]
                # renormalize
                self.all_f_t2p_preproc =self.zscore_all_f_t2p(self.all_f_t2p_preproc)
            if rem_zero_rows:
                # get zero rows in any of the datasets
                zero_rows = np.any([np.sum(np.isnan(f), axis=1) for f in self.all_f_t2p_preproc], axis=0)
                print(f'Number of zero rows: {np.sum(zero_rows)}') 
                self.all_f_t2p_preproc= [f[~zero_rows, :] for f in self.all_f_t2p_preproc]

                
        def zscore(self, f, axis=1): 
            '''this method calculates the z-score for each element in the input array f along the specified axis, ignoring NaN values.'''
            return (f - np.nanmean(f, axis=axis, keepdims=True)) / np.nanstd(f, axis=axis, keepdims=True)

        def zscore_all_f_t2p(self, all_f_t2p):
            return [zscore(f, axis=1) for f in all_f_t2p] #axis=1 means that we are normalizing each neuron's activity (each line of the matrix f)
        
                
        def plot_track2p_rasters(self, all_f_t2p, bin_size=1, vmin=None, vmax=None):
                
            fig, ax = plt.subplots(len(all_f_t2p), 1, figsize=(6, len(all_f_t2p)*1), dpi=150) #rows, columns, size, resolution

            for i, f in enumerate(all_f_t2p):
                ax[i].imshow(zscore(f, axis=1), aspect='auto', cmap='Greys', vmin=vmin, vmax=vmax)
                # only first and last yticks
                ax[i].set_yticks([0, f.shape[0]-1])
                ax[i].set_yticklabels([0, f.shape[0]])
                # move the text of the yticklabels a bit: the first one down and the last one up
                ax[i].set_yticklabels(ax[i].get_yticklabels(), rotation=0, va='center', ha='right')
        
                # only first and last xticks on y axis in minutes
                if i == len(all_f_t2p)-1:
                    ax[i].set_xlabel('Frames')
                    ax[i].set_xticks([0, f.shape[1]/2, f.shape[1]-1])
                    ax[i].set_xticklabels([0, int((f.shape[1]*bin_size)/2), int(f.shape[1]*bin_size)])
            
                else:
                    ax[i].set_xticks([])
                    ax[i].set_xticklabels([])
            
                plt.subplots_adjust(bottom=0.2)
            buf = io.BytesIO()
            plt.savefig(buf, format='png')
            buf.seek(0)
            pixmap = QPixmap()
            pixmap.loadFromData(buf.getvalue())
            scene=QGraphicsScene()
            scene.addPixmap(pixmap)
            self.view.setScene(scene)
            print('Done')
                
            

            
        def concatenate_rasters(self):

                track_ops=self.main_window.central_widget.data_management.track_ops
                t2p_folder_path= os.path.dirname(track_ops.all_ds_path[0])
                print(t2p_folder_path)
                if track_ops.nplanes > 1:
                    self.results_by_plane = {}
                    print(track_ops.nplanes)
                    for plane in range (track_ops.nplanes):
                        print(plane)
                        if plane == self.main_window.central_widget.data_management.plane:
                            print('plane is equal to the current plane, skipping to the next plane')
                            self.results_by_plane[plane]={
                            'all_ft2p': self.main_window.central_widget.data_management.all_f_t2p,
                            'all_fneu2p': self.main_window.central_widget.data_management.all_fneu
                                        }
                            continue
                        print('plane is not equal to the current plane')
                        t2p_match_mat = np.load(os.path.join(t2p_folder_path,"track2p" ,f"plane{plane}_match_mat.npy"), allow_pickle=True)
                        t2p_match_mat_allday = t2p_match_mat[~np.any(t2p_match_mat == None, axis=1), :] 
                        trace_type=self.main_window.central_widget.data_management.trace_type #common to all planes
                        print(f"Processing plane {plane}")
                        self.process_plane(plane,track_ops,t2p_match_mat_allday,trace_type)
                    for plane, data in self.results_by_plane .items():
                            print(f"Plane {plane}:")
                            print(f"  all_ft2p: {len(data['all_ft2p'])}")
                            print(f"  {len(data['all_ft2p'][0])}")
                   
                    # Initialiser une liste pour stocker les éléments concaténés
                    concatenated_elements = []
                    num_elements = len(self.results_by_plane[0]['all_ft2p'])
                    print(f"Number of elements: {num_elements}")
                    # Itérer sur les indices des éléments
                    for i in range(num_elements):
                        elements_to_concatenate = []
                        for plane in range(track_ops.nplanes):
                            if 'all_ft2p' in self.results_by_plane[plane] and isinstance(self.results_by_plane[plane]['all_ft2p'], list):
                            # Récupérer l'élément i de 'all_ft2p' pour le plan actuel
                                element = self.results_by_plane[plane]['all_ft2p'][i]
                                elements_to_concatenate.append(element)
                            else:
                                print("La clé 'all_ft2p' n'existe pas ou n'est pas une liste.")
                        if elements_to_concatenate:
                            concatenated_element = np.vstack(elements_to_concatenate)
                            concatenated_elements.append(concatenated_element)

                    # Afficher les éléments concaténés
                    for idx, concatenated_element in enumerate(concatenated_elements):
                        print(f"Concatenated element {idx}:")
                        #print(concatenated_element)
                        print(concatenated_element.shape)
                    
                    return concatenated_elements



        def process_plane(self, plane, track_ops, t2p_match_mat_allday, trace_type):
            all_ft2p=[]
            all_fneu2p=None
            for (i, ds_path) in enumerate(track_ops.all_ds_path):
                iscell = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'iscell.npy'), allow_pickle=True)
                if trace_type == 'F' :
                    print('F trace')
                    f = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'F.npy'), allow_pickle=True)
                if trace_type == 'spks':
                    print('spks trace')
                    f = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'spks.npy'), allow_pickle=True)
                if trace_type == 'dF/F0':
                    print('dF/F0 trace')
                    if all_fneu2p is None:
                        all_fneu2p= []
                    f = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'F.npy'), allow_pickle=True)
                    fneu = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'Fneu.npy'), allow_pickle=True)
                    if track_ops.iscell_thr is None:
                        fneu_iscell = fneu[iscell[:, 0] == 1, :]
                    else:
                        fneu_iscell = fneu[iscell[:, 1] > track_ops.iscell_thr, :]
                    fneu_t2p= fneu_iscell[t2p_match_mat_allday[:, i].astype(int), :]
                    all_fneu2p.append(fneu_t2p)
                if track_ops.iscell_thr is None:
                    f_iscell = f[iscell[:, 0] == 1, :]
                else:
                    f_iscell = f[iscell[:, 1] > track_ops.iscell_thr, :] 
                
                f_t2p = f_iscell[t2p_match_mat_allday[:, i].astype(int), :]
                all_ft2p.append(f_t2p)

            self.results_by_plane[plane]={
                    'all_ft2p': all_ft2p,
                    'all_fneu2p': all_fneu2p
                                        }
                


import numpy as np 
from skimage.filters import threshold_otsu, threshold_minimum
from scipy.optimize import linear_sum_assignment

from track2p.match.utils import get_cost_mat, get_iou, init_all_pl_match_mat

# assigment of ROIs in each ref-reg pair

def get_all_ds_assign(track_ops, all_ds_all_roi_ref, all_ds_all_roi_reg):

    all_ds_assign = []
    all_ds_assign_thr = []
    all_ds_thr_met = []
    all_ds_thr = []

    for i in range(len(track_ops.all_ds_path)-1):
        print(f'Finding matches in ref-reg pair: {i+1}/{len(track_ops.all_ds_path)-1}')
        ds_assign = []
        ds_assign_thr = []
        ds_thr_met = []
        ds_thr = []
        for j in range(track_ops.nplanes):
            all_roi_ref = all_ds_all_roi_ref[i][j]
            all_roi_reg = all_ds_all_roi_reg[i][j]

            # 1) compute cost matrix (currently two methods available, see DefaultTrackOps)
            cost_mat, all_inds_ref_filt, all_inds_reg_filt = get_cost_mat(all_roi_ref, all_roi_reg, track_ops) 

            # 2) optimally assign pairs
            ref_ind_filt, reg_ind_filt = linear_sum_assignment(cost_mat)

            # 3) convert them to pre-filtered indices (these are the indices of the ROIs after iscell)
            ref_ind = all_inds_ref_filt[ref_ind_filt]
            reg_ind = all_inds_reg_filt[reg_ind_filt]

            # 4) for each matched pair (len(all_roi_ref)) compute thresholding metric (in this case IOU, the filtering will be done afterwards in the all-day assignment)
            thr_met = get_iou(all_roi_ref[:,:,ref_ind], all_roi_reg[:,:,reg_ind])
            thr_met_compute = thr_met[thr_met>0] if track_ops.thr_remove_zeros else thr_met # remove zeros for computing the threshold (otsu thresholding is squed 

            # 5) compute otsu threshold on thr_met
            if track_ops.thr_method == 'otsu':
                thr = threshold_otsu(thr_met_compute)
            elif track_ops.thr_method == 'min':
                thr = threshold_minimum(thr_met_compute)


            ds_assign.append([ref_ind, reg_ind])
            ds_assign_thr.append([ref_ind[thr_met>thr], reg_ind[thr_met>thr]])
            ds_thr_met.append(thr_met)
            ds_thr.append(thr)

        all_ds_assign.append(ds_assign)
        all_ds_assign_thr.append(ds_assign_thr)
        all_ds_thr_met.append(ds_thr_met)
        all_ds_thr.append(ds_thr)
        print(f'Done ref-reg pair: {i+1}/{len(track_ops.all_ds_path)-1}')

    return all_ds_assign, all_ds_assign_thr, all_ds_thr_met, all_ds_thr


# propagating matches across all days

def get_all_pl_match_mat(all_ds_all_roi_ref, all_ds_assign_thr, track_ops):

    all_pl_match_mat = init_all_pl_match_mat(all_ds_all_roi_ref, all_ds_assign_thr, track_ops)

    for i in range(track_ops.nplanes):
        pl_match_mat = all_pl_match_mat[i]
        # now for each row in the match matrix (each ROI in the ref recording) we need to find the match across all days, if there is none then we leave it as None

        for roi_idx in range(pl_match_mat.shape[0]): # roi_idx is the index on first session
            # if first column is none then we skip this row
            if pl_match_mat[roi_idx, 0] is None:
                continue
            # otherwise we find the match in the all_ds_assign_thr
            else:
                ref_roi_ds0 = pl_match_mat[roi_idx, 0]
                track_roi = np.array(ref_roi_ds0)
                for ds_ind in range(pl_match_mat.shape[1]-1):
                    matches = all_ds_assign_thr[ds_ind][i]
                    
                    ref_ind = matches[0]
                    reg_ind = matches[1]
                    
                    reg_ind_ind = np.where(ref_ind==track_roi.item())[0]

                    # if there is a match then we update the track_roi
                    if reg_ind_ind.size>0:
                        track_roi = reg_ind[reg_ind_ind]
                        pl_match_mat[roi_idx, ds_ind+1] = track_roi.item()
                        
                    # if there is no match then we stop tracking this ROI
                    else:
                        break
                    
        # compute how many ROIs are tracked across all days
        n_tracked = np.sum(np.all(pl_match_mat!=None, axis=1))
        print(f'Number of ROIs tracked in plane{i} across all days: {n_tracked}')
        track_ops.all_pl_match_mat = all_pl_match_mat
        track_ops.n_tracked = n_tracked
        
    return all_pl_match_mat


import numpy as np
from scipy.spatial.distance import cdist
from skimage import measure
from skimage.filters import threshold_otsu

# compute centroids
def get_centroids(all_roi):
    centroids = []
    for i in range(all_roi.shape[2]):
        roi = all_roi[:,:,i]
        labels = measure.label(roi)
        features = measure.regionprops(labels)
        try:
            centroids.append(features[0].centroid)
        except IndexError:
            centroids.append([0, 0])

    return np.array(centroids)

def get_cent_dist_mat(all_roi_ref, all_roi_reg):
    # compute distances
    centroids_ref = get_centroids(all_roi_ref)
    centroids_reg = get_centroids(all_roi_reg)
    distances = cdist(centroids_ref, centroids_reg)
    return distances

def filt_non_overlap(all_roi1, all_roi2, cent_dist_mat):
    all_inds = np.arange(all_roi1.shape[2])
    filt_inds_ref = []
    for i in range(all_roi1.shape[2]):
        # get the index of closest roi from the distances matrix
        roi = all_roi1[:,:,i]
        closest_roi_idx = np.argmin(cent_dist_mat[i,:])
        closest_roi = all_roi2[:,:,closest_roi_idx]
        intersection = roi*closest_roi
        if np.sum(intersection)==0: # TODO:maybe it overlaps a bit with non-first closest roi?
            filt_inds_ref.append(i)

    # get all indices that are not in filt_inds_ref
    all_inds_filt = all_inds[~np.isin(all_inds, filt_inds_ref)]  
    return all_inds_filt

def get_cent_dist_mat_non_overlap(all_roi_ref, all_roi_reg):

    cent_dist_mat = get_cent_dist_mat(all_roi_ref, all_roi_reg)
    
    # get indices of neuorns that overlap with their closest neighbor (at least will have some chance to match)
    all_inds_ref_filt = filt_non_overlap(all_roi_ref, all_roi_reg, cent_dist_mat)
    all_inds_reg_filt = filt_non_overlap(all_roi_reg, all_roi_ref, cent_dist_mat.T)

    cost_mat = cent_dist_mat[all_inds_ref_filt, :]
    cost_mat = cost_mat[:, all_inds_reg_filt]

    return cost_mat, all_inds_ref_filt, all_inds_reg_filt

def get_cost_mat(all_roi_ref, all_roi_reg, track_ops):
    # compute distances
    if track_ops.matching_method=='cent': # simple assignment based on centroids (probably works with sparse data but not in development)
        cost_mat = get_cent_dist_mat(all_roi_ref, all_roi_reg)
        all_inds_ref_filt = np.arange(all_roi_ref.shape[2]) # here we don't filter the indices
        all_inds_reg_filt = np.arange(all_roi_reg.shape[2]) # here we don't filter the indices
    elif track_ops.matching_method=='cent_int-filt': # here to simplify the matching we first filter out the ROIs that have no intersection with their closest neighbor
        # this also outputs the indices of neurons after filtering
        # costa mat here is smaller than above since the matches are filtered
        # additionally when outputting we need to be careful to index with the filtered indices as well
        cost_mat, all_inds_ref_filt, all_inds_reg_filt = get_cent_dist_mat_non_overlap(all_roi_ref, all_roi_reg) 
    elif track_ops.matching_method=='iou':
        cost_mat = 1-get_cross_iou_mat(all_roi_ref, all_roi_reg, dist_thr=track_ops.iou_dist_thr)
        all_inds_ref_filt = np.arange(all_roi_ref.shape[2])
        all_inds_reg_filt = np.arange(all_roi_reg.shape[2])
    else:
        raise Exception('Matching method not implemented')
    
    print(f'cost_mat computed with method: {track_ops.matching_method}')
    print(f'cost_mat shape: {cost_mat.shape}')
    print(f'cost_mat min: {np.min(cost_mat)}')
    print(f'cost_mat max: {np.max(cost_mat)}')
    
    return cost_mat, all_inds_ref_filt, all_inds_reg_filt

def get_iou(all_roi_ref, all_roi_reg):

    ious = []
    for i in range(all_roi_ref.shape[2]):
        roi_ref = all_roi_ref[:,:,i]
        roi_reg = all_roi_reg[:,:,i]
        intersection = np.sum(np.logical_and(roi_ref, roi_reg))
        union = np.sum(np.logical_or(roi_ref, roi_reg))
        ious.append(intersection/union)

    return np.array(ious)

def get_cross_iou_mat(all_roi_ref, all_roi_reg, dist_thr=16):
    # if the distance between two rois is larger than dist_thr, we assume they are not the same cell and just skip the computation
    distances = get_cent_dist_mat(all_roi_ref, all_roi_reg)

    cross_iou_mat = np.zeros((all_roi_ref.shape[2], all_roi_reg.shape[2]))
    for i in range(all_roi_ref.shape[2]):
        for j in range(all_roi_reg.shape[2]):
            if distances[i,j] > dist_thr: # skipping if far apart
                continue
            # compute IOU
            intersection = np.logical_and(all_roi_ref[:,:,i], all_roi_reg[:,:,j])
            union = np.logical_or(all_roi_ref[:,:,i], all_roi_reg[:,:,j])
            iou_score = np.sum(intersection) / np.sum(union)
            cross_iou_mat[i,j] = iou_score

    return cross_iou_mat

def init_all_pl_match_mat(all_ds_all_roi_ref, all_ds_assign_thr, track_ops):
    all_pl_match_mat = []
    for i in range(track_ops.nplanes):
        # set up the match matrix for each plane it will be size of all iscell ROIs in the ref recording x number of datasets
        pl_match_mat = np.full((all_ds_all_roi_ref[0][i].shape[2], len(track_ops.all_ds_path)), None)
        all_pl_match_mat.append(pl_match_mat)
    # populate first row of the match matrices with the matches from the first ref-reg pair
    for i in range(track_ops.nplanes):
        pl_match_mat = all_pl_match_mat[i]
        assign_thr = all_ds_assign_thr[0][i] # zeroth dataset, ith plane
        ref_ind = assign_thr[0]
        pl_match_mat[ref_ind, 0] = ref_ind

    return all_pl_match_mat

def filt_by_otsu(vect_filt, vect_comp):
    # vect_filt is the vector that we want to filter
    # vect_comp is the vector that we want to compute the otsu threshold on

    thresh = threshold_otsu(vect_comp)
    return vect_filt[vect_comp>thresh]


import os
import numpy as np

def check_nplanes(track_ops):
    all_nplanes = []

    for ds_path in track_ops.all_ds_path:
        # check how many subfolders starting with plane* in suite2p folder
        n_planes = len([name for name in os.listdir(ds_path + '/suite2p') if name.startswith('plane')])
        print(f'Found {n_planes} planes in {ds_path}')
        all_nplanes.append(n_planes)
    track_ops.all_nplanes = all_nplanes
    # if all elements in all_n_planes are the same, then save it in track_ops.n_planes
    if all_nplanes.count(all_nplanes[0]) == len(all_nplanes):
        track_ops.nplanes = all_nplanes[0]
        print(f'Found {track_ops.nplanes} planes in all datasets')
    else:
        print('Found different number of planes in different datasets')
        print('Please check your dataset paths')
        print('Exiting...')
        exit()


# loads mean images
def load_all_imgs(track_ops):
    all_ds_avg_ch1 = []
    all_ds_avg_ch2 = []
    all_ds_nchannels = []

    for ds_path in track_ops.all_ds_path:
        ds_nchannels = []
        ds_avg_ch1 = []
        ds_avg_ch2 = []

        for i in range(track_ops.nplanes):
            ops = np.load(ds_path + '/suite2p/plane' + str(i) + '/ops.npy', allow_pickle=True).item()
            nchannels = ops['nchannels']
            print('nchannels: ' + str(nchannels) + ' for plane ' + str(i) + ' in dataset ' + ds_path)
            ds_avg_ch1.append(ops['meanImg'])
            ds_avg_ch2.append(ops['meanImg_chan2']) if nchannels==2 else ds_avg_ch2.append(None)
            ds_nchannels.append(nchannels)

        all_ds_avg_ch1.append(ds_avg_ch1)
        all_ds_avg_ch2.append(ds_avg_ch2)
        all_ds_nchannels.append(ds_nchannels)

    track_ops.all_ds_avg_ch1 = all_ds_avg_ch1
    track_ops.all_ds_avg_ch2 = all_ds_avg_ch2
    track_ops.all_ds_nchannels = all_ds_nchannels

    # if all elements in all_ds_nchannels are the same, then print its fine otherwise exit
    if all_ds_nchannels.count(all_ds_nchannels[0]) == len(all_ds_nchannels):
        track_ops.nchannels = all_ds_nchannels[0][0]
        print(f'Found {track_ops.nchannels} channels in all datasets')
    else:    
        print('Found different number of channels in different datasets')
        print('Please check your dataset paths')
        print('Exiting...')
        exit()

    return all_ds_avg_ch1, all_ds_avg_ch2

def load_all_ds_stat_iscell(track_ops):
    all_ds_stat_iscell = []
    for (i, ds_path) in enumerate(track_ops.all_ds_path):
        ds_stat_iscell = []
        for j in range(track_ops.nplanes):
            stat = np.load(os.path.join(ds_path, 'suite2p', f'plane{j}', 'stat.npy'), allow_pickle=True)
            iscell = np.load(os.path.join(ds_path, 'suite2p', f'plane{j}', 'iscell.npy'), allow_pickle=True)
            if track_ops.iscell_thr==None:
                stat_iscell = stat[iscell[:,0]==1]
            else: 
                stat_iscell = stat[iscell[:,1]>track_ops.iscell_thr]
            ds_stat_iscell.append(stat_iscell)
        all_ds_stat_iscell.append(ds_stat_iscell)

    return all_ds_stat_iscell

def load_all_ds_ops(track_ops):
    all_ds_ops = []
    for ds_path in track_ops.all_ds_path:
        ds_ops = []
        for j in range(track_ops.nplanes):
            ops = np.load(os.path.join(ds_path, 'suite2p', f'plane{j}', 'ops.npy'), allow_pickle=True).item()
            ds_ops.append(ops)
        all_ds_ops.append(ds_ops)
    
    return all_ds_ops

def load_all_ds_mean_img(track_ops, ch=1):
    all_ds_ops = load_all_ds_ops(track_ops)
    all_ds_mean_img = [] 
    for ds_ops in all_ds_ops:
        ds_mean_img = []
        for ops in ds_ops:
            mean_img = ops['meanImg'] if ch==1 else ops['meanImg_chan2']
            ds_mean_img.append(mean_img)
        all_ds_mean_img.append(ds_mean_img)
        
    return all_ds_mean_img

def load_all_ds_centroids(all_ds_stat_iscell, track_ops):
    all_ds_centroids = []
    for i in range(len(track_ops.all_ds_path)):
        ds_centroids = []
        for stat_iscell in all_ds_stat_iscell[i]:
            centroids = []
            for roi_stat in stat_iscell:
                centroids.append(roi_stat['med'])
            ds_centroids.append(np.array(centroids))
        all_ds_centroids.append(ds_centroids)
        
    return all_ds_centroids



# make dummy track_ops object (would be input from command line or gui)
import os
from track2p.io.utils import make_dir

class DefaultTrackOps:
    def __init__(self):
        # input list of dataset paths (each contains a 'suite2p' folder)
        self.all_ds_path = [
            'data/ac/ac444118/2022-09-14_a',
            'data/ac/ac444118/2022-09-15_a',
            'data/ac/ac444118/2022-09-16_a'
        ]
 
        self.save_path = 'data/ac/ac444118/track2p/'
        # self.save_path = 'data/jm/jm032/track2p/'
        self.reg_chan = 0 # channel to use for registration (0=functional, 1=anatomical) (1 is not always available)
        self.transform_type = 'affine' # 'affine' or 'nonrigid'
        self.iscell_thr = 0.50 # threshold for iscell.npy (only keep ROIs with iscell > iscell_thr) (here lowering this can be good and non-detrimental -> artefacts are unlikely to be consistently present in all datasets)

        self.matching_method='iou' # 'iou', 'cent' or 'cent_int-filt'  (iou takes longer but is more accurate, cent is faster but less accurate)
        self.iou_dist_thr = 16 # distance between centroids (in pixels) above which to skip iou computation (to save time) (this is only relevant if self.matching_method=='iou')

        self.thr_remove_zeros = False # remove zeros from thr_met before computing automatic threshold (this is useful when there are many zeros in thr_met, which can skew the thresholding)
        self.thr_method = 'min' # 'otsu' or 'min' (min is just local minimum of pdf of thr_met)

        # do not change these
        self.show_roi_reg_output = False # this is slow because plt.contour is slow and also very memory intensive(it can easily crash) but the visualisation is nice for presentations (for example by increasing self.iscell_thr)

        # plotting parameters
        self.win_size = 48 # window size for visualising matched ROIs across days (crop of mean image)
        self.sat_perc = 99.9 # percentile to saturate image at (only affects visualisation not the registration/matching)
        
        
        self.colors = None 
     
        #self.vector_curation_plane_0 = None
        #self.vector_curation_plane_1 = None
        
        self.save_in_s2p_format = False # save the output in suite2p format (this is useful for downstream analysis with suite2p)

        # make the output directories when initialising the object
        
    def init_save_paths(self):
        self.save_path = os.path.join(self.save_path, 'track2p/')
        self.save_path=self.save_path.replace("\\", "/")
        self.save_path_fig = os.path.join(self.save_path, 'fig/')
        self.save_path_fig=self.save_path_fig.replace("\\", "/")
        make_dir(self.save_path)
        make_dir(self.save_path_fig)


    def to_dict(self):
        # this is useful for saving the object to avoid needing class definition in downstream analysis
        track_ops_dict = {}
        for attr in dir(self):
            if not attr.startswith('__') and not callable(getattr(self, attr)):
                track_ops_dict[attr] = getattr(self, attr)
        return track_ops_dict

    def from_dict(self, track_ops_dict):
        # loop through all the keys and set the attributes
        for key in track_ops_dict:
            setattr(self, key, track_ops_dict[key])



import os
import numpy as np
import matplotlib.colors as mcolors
import random
from types import SimpleNamespace
from scipy.ndimage import maximum_filter1d, minimum_filter1d, gaussian_filter

class DataManagement:
    def __init__(self, central_widget):
        self.central_widget = central_widget
        self.main_window = central_widget.main_window
        self.all_f_t2p = []
        self.all_ops = []
        self.all_stat_t2p = []
        self.all_iscell = []
        self.all_fneu= []
        self.colors = None
        self.t2p_match_mat_allday = None
        self.track_ops = None
        self.vector_curation_t2p = None
        self.curation_npy = None
        self.colors_copy=None
        self.plane=None
        self.trace_type=None
        
    def reset_attributes(self):
        self.all_f_t2p = []
        self.all_ops = []
        self.all_stat_t2p = []
        self.all_iscell = []
        self.all_fneu= []
        self.colors = None
        self.t2p_match_mat_allday = None
        self.track_ops = None
        self.vector_curation_t2p = None
        
    def import_files(self, t2p_folder_path, plane, trace_type, channel):
        
        self.plane=plane
        self.trace_type=trace_type

        if self.central_widget.fluorescences_plotting is not None:
            self.central_widget.clear()
        # load track2p outputs           
        t2p_match_mat = np.load(os.path.join(t2p_folder_path,"track2p" ,f"plane{plane}_match_mat.npy"), allow_pickle=True)
        self.t2p_match_mat_allday = t2p_match_mat[~np.any(t2p_match_mat == None, axis=1), :] # remove rows with None values
        track_ops_dict = np.load(os.path.join(t2p_folder_path, "track2p", "track_ops.npy"), allow_pickle=True).item()
        track_ops = SimpleNamespace(**track_ops_dict)
      
        
        # process suite2p files 
        for (i, ds_path) in enumerate(track_ops.all_ds_path):
            ops = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'ops.npy'), allow_pickle=True).item()
            stat = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'stat.npy'), allow_pickle=True)
            iscell = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'iscell.npy'), allow_pickle=True)
            if trace_type == 'F' :
                print('F trace')
                f = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'F.npy'), allow_pickle=True)
            if trace_type == 'spks':
                print('spks trace')
                f = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'spks.npy'), allow_pickle=True)
            if trace_type == 'dF/F0':
                print('dF/F0 trace')
                f = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'F.npy'), allow_pickle=True)
                fneu = np.load(os.path.join(ds_path, 'suite2p', f'plane{plane}', 'Fneu.npy'), allow_pickle=True)
                if track_ops.iscell_thr is None:
                    fneu_iscell = fneu[iscell[:, 0] == 1, :]
                else:
                    fneu_iscell = fneu[iscell[:, 1] > track_ops.iscell_thr, :]
                fneu_t2p= fneu_iscell[self.t2p_match_mat_allday[:, i].astype(int), :]
                self.all_fneu.append(fneu_t2p)
            if track_ops.iscell_thr is None:
                    stat_iscell = stat[iscell[:, 0] == 1]
                    f_iscell = f[iscell[:, 0] == 1, :]
            else:
                    stat_iscell = stat[iscell[:, 1] > track_ops.iscell_thr]
                    f_iscell = f[iscell[:, 1] > track_ops.iscell_thr, :] 
               

            stat_t2p = stat_iscell[self.t2p_match_mat_allday[:, i].astype(int)]
            f_t2p = f_iscell[self.t2p_match_mat_allday[:, i].astype(int), :]
            self.all_stat_t2p.append(stat_t2p)
            self.all_f_t2p.append(f_t2p)
            self.all_ops.append(ops)
            self.all_iscell.append(iscell)

        if trace_type == 'dF/F0':
            for i in range(len(self.all_f_t2p)):
                f_prc=self.F_processing(F=self.all_f_t2p[i], Fneu=self.all_fneu[i], fs=self.all_ops[i]['fs'])
                self.all_f_t2p[i]=f_prc

       
        attr_name = 'vector_curation_plane_' + str(plane)
        if hasattr(track_ops, attr_name):
            key = 'vector_curation_plane_' + str(plane)
            self.vector_curation_t2p=track_ops_dict[key] 
        else:
            vector_curation_keys=np.arange(self.t2p_match_mat_allday.shape[0]) 
            vector_curation_values = np.ones_like(vector_curation_keys)
            self.vector_curation_t2p_dict = dict(zip(vector_curation_keys, vector_curation_values))
            values = list(self.vector_curation_t2p_dict.values())
            self.vector_curation_t2p = np.array(values)
            key = 'vector_curation_plane_' + str(plane)
            track_ops_dict[key] = self.vector_curation_t2p 
            np.save(os.path.join(t2p_folder_path, "track2p", "track_ops.npy"), track_ops_dict) 

  
        attr_name_color = 'colors_plane_' + str(plane)
        if hasattr(track_ops, attr_name_color):
            self.colors= track_ops_dict[attr_name_color]
        else:
            self.colors=self.generate_vibrant_colors(len(self.all_stat_t2p[0]))
            track_ops_dict[attr_name_color] = self.colors
            np.save(os.path.join(t2p_folder_path, "track2p", "track_ops.npy"), track_ops_dict) 
           
            
        self.track_ops = track_ops
        track_ops_dict = np.load(os.path.join(t2p_folder_path, "track2p", "track_ops.npy"), allow_pickle=True).item()
        self.central_widget.track_ops_dict=track_ops_dict
        self.main_window.status_bar.vector_curation_t2p = self.vector_curation_t2p
        
        self.main_window.status_bar.vector_curation_t2p = self.vector_curation_t2p
            
        self.main_window.status_bar.spin_box.setSuffix(f'/{len(self.t2p_match_mat_allday)-1}')
        self.main_window.status_bar.spin_box.setMinimum(0)
        self.main_window.status_bar.spin_box.setMaximum(len(self.t2p_match_mat_allday)-1) 
        
        num_ones = {}  
        
        for cell, line in enumerate(t2p_match_mat): 
                all_iscell_value=[]
                for day,index_match in enumerate(line):
                    if track_ops.iscell_thr is None:
                        if index_match is None:
                            all_iscell_value.append(0)
                        else: 
                            iscell=self.all_iscell[day]
                            indices_lignes_1 = np.where(iscell[:,0]==1)[0] # take the indices where the ROIs were considered as cells in suite2p
                            true_index=indices_lignes_1[index_match] # take the "true index"
                            iscell_value=iscell[true_index,0] 
                            all_iscell_value.append(iscell_value)
                    else: 
                        if index_match is None:
                            all_iscell_value.append(0)
                        else: 
                            iscell=self.all_iscell[day]
                            indices_lignes_1= np.where(iscell[:,1]>track_ops.iscell_thr)[0] # take the indices where the ROIs have a probability greater than trackops.is_cell_thr
                            true_index=indices_lignes_1[index_match] # take the "true index"
                            iscell_value=iscell[true_index,0] 
                            all_iscell_value.append(iscell_value)
                num_ones[cell] = all_iscell_value.count(1)

       
        for day in range(len(self.t2p_match_mat_allday[1]) + 1):
            count_cells_day=0
            for value in num_ones.values():
                if value == day:
                    count_cells_day+=1
            print(f'Number of cells present over {day} day(s): {count_cells_day}')
                
        for i, line in enumerate(self.t2p_match_mat_allday): 
            if self.vector_curation_t2p[i]==0:
                self.colors[i] =(0.78, 0.78, 0.78)   
                print(f'ROI {i} has been considered as "not cell"')  
        
      
        self.central_widget.create_mean_img(channel)
        self.central_widget.vector_curation_t2p = self.vector_curation_t2p
        self.central_widget.display_first_ROI(0)

        
        
    def generate_vibrant_colors(self, num_colors):
        vibrant_colors = []
        for _ in range(num_colors):
            l = np.random.uniform(0.55, 0.80) 
            color = mcolors.hsv_to_rgb((random.random(), 1, l)) 
            vibrant_colors.append(color)

        return vibrant_colors
    
    
    def F_processing(self,F, Fneu, fs, neucoeff=0.0, baseline='maximin', sig_baseline=10.0, win_baseline=60.0, prctile_baseline: float = 8):
    
        print(F.shape)
        print(Fneu.shape)
    #neuropil substraction 
        Fc = F - neucoeff * Fneu

    # baseline operation  
        win = int(win_baseline * fs)
        if baseline == "maximin":
            Flow = gaussian_filter(Fc, [0., sig_baseline])
            Flow = minimum_filter1d(Flow, win)
            Flow = maximum_filter1d(Flow, win)
        elif baseline == "constant":
            Flow = gaussian_filter(Fc, [0., sig_baseline])
            Flow = np.amin(Flow)
        elif baseline == "constant_prctile":
            Flow = np.percentile(Fc, prctile_baseline, axis=1)
            Flow = np.expand_dims(Flow, axis=1)
        else:
            Flow = 0.

        F = Fc - Flow
        
        print('DONE')

        return F
      



from track2p.gui.window_manager import WindowManager
from track2p.gui.toolbar import Toolbar
from track2p.gui.statusbar import StatusBar
from track2p.gui.data_management import DataManagement
from track2p.gui.central_widget import CentralWidget
from PyQt5.QtWidgets import QApplication,QMainWindow
from PyQt5.QtGui import QIcon

class MainWindow(QMainWindow):
    
    def __init__(self):
        super(MainWindow, self).__init__()
        self.main_window = self
        self.window_manager = WindowManager(self)
        self.data_management = DataManagement(self)
        self.central_widget = CentralWidget(self)
        self.toolbar = Toolbar(self)
        self.status_bar = StatusBar(self)
   
        
        
        self.initUI()

    def initUI(self):
        
        
        self.setStyleSheet(
            "QTabWidget::pane { border: 1px solid #666; }"
            "QTabWidget::tab-bar { alignment: center; }"
            "QTabBar::tab { background-color: #666; color: white; }"
            "QTabBar::tab:selected { background-color: #222; color: white; }"
            "QSplitter::handle { background: #888; }"
            "QFrame { background-color: black; color: black;  border: 1px solid black;}"
            "QLabel { color: black; background-color: none; border: none; font-size: 13px}"
            "QPushButton { background-color: #666; color: white; border: 1px solid #888; }"
            "QPushButton:hover { background-color: #888; color: white; }"
            "QPushButton:pressed { background-color: #333; color: white; }"
            "QToolButton:pressed { background-color: #888; }"
            "QComboBox { background-color: black; color: white; }"
            "QComboBox QAbstractItemView { background-color: #666; color: white; }"
            )
        
        self.setWindowTitle("track2p GUI")

        self.setCentralWidget(self.central_widget)
        self.addToolBar(self.toolbar)
        self.setStatusBar(self.status_bar)
        QApplication.setStyle('Cleanlooks')
        self.showMaximized()
        


        


            



# for now the only algorithm is elastix, TODO: add other algorithms and run them in the same way within this loop

from track2p.register.elastix import reg_img_elastix
from track2p.io.loaders import load_stat_ds_plane, get_all_roi_array_from_stat
from track2p.register.elastix import itk_reg_all_roi

def run_reg_loop(all_ds_ref_img, all_ds_mov_img, track_ops):
    all_ds_mov_img_reg = []
    all_ds_reg_params = []

    for (i, ds_ref_img) in enumerate(all_ds_ref_img):
        ds_mov_img = all_ds_mov_img[i]
        ds_mov_img_reg = []
        ds_reg_params = []

        for j in range(track_ops.nplanes):
            ref_img = ds_ref_img[j]
            mov_img = ds_mov_img[j]
            mov_img_reg,reg_params = reg_img_elastix(ref_img, mov_img, track_ops)
            ds_mov_img_reg.append(mov_img_reg)
            ds_reg_params.append(reg_params)

        all_ds_mov_img_reg.append(ds_mov_img_reg)
        all_ds_reg_params.append(ds_reg_params)
        
    track_ops.all_ds_mov_img_reg = all_ds_mov_img_reg
    # track_ops.all_ds_reg_params = all_ds_reg_params # for now not saving elastix transform object (TODO: transforme it to serializable to be able to pickle (for example by saving params to a dictionary))
    
    return all_ds_mov_img_reg, all_ds_reg_params


def reg_all_ds_all_roi(all_ds_reg_params, track_ops):
    all_ds_all_roi_array_ref = []
    all_ds_all_roi_array_mov = []
    all_ds_all_roi_array_reg = []
    all_ds_roi_counter = [] # this will keep track of how many cells make it thorugh the registration

    for i in range(len(track_ops.all_ds_path)-1):
        print(f'...\nTransforming ROIs for registration {i}/{len(track_ops.all_ds_path)-1}')

        ds_all_roi_array_ref = [] # for one dataset (all planes)
        ds_all_roi_array_mov = []
        ds_all_roi_array_reg = []
        ds_roi_counter_ref = []
        ds_roi_counter_mov = []

        for j in range(track_ops.nplanes):

            # 1) Set paths and transformation
            ref_ds_path = track_ops.all_ds_path[i]
            reg_ds_path = track_ops.all_ds_path[i+1]

            reg_params = all_ds_reg_params[i][j]

            # 2) Load ROIs # TODO: add (non-s2p dependent) Cellpose ROIs compatibility
            stat_ref, roi_counter_ref = load_stat_ds_plane(ref_ds_path, track_ops, plane_idx=j) # loading rois for one dataset one plane
            stat_mov, roi_counter_mov = load_stat_ds_plane(reg_ds_path, track_ops, plane_idx=j) # loading rois for one dataset one plane

            all_roi_array_ref = get_all_roi_array_from_stat(stat_ref, track_ops) # for dataset i, plane j
            all_roi_array_mov = get_all_roi_array_from_stat(stat_mov, track_ops) # for dataset i, plane j

            # 3) Apply transformation
            all_roi_array_reg = itk_reg_all_roi(all_roi_array_mov, reg_params)

            # 4) append
            ds_all_roi_array_ref.append(all_roi_array_ref)
            ds_all_roi_array_mov.append(all_roi_array_mov)
            ds_all_roi_array_reg.append(all_roi_array_reg)

            ds_roi_counter_ref.append(roi_counter_ref)
            ds_roi_counter_mov.append(roi_counter_mov) # only keep this one if it is the last pair
        
        print('Done with dataset...')

        all_ds_all_roi_array_ref.append(ds_all_roi_array_ref)
        all_ds_all_roi_array_mov.append(ds_all_roi_array_mov)
        all_ds_all_roi_array_reg.append(ds_all_roi_array_reg)
        
        all_ds_roi_counter.append(ds_roi_counter_ref) # keep both if its last pair (last recording does't appear as a ref)
        if i == len(track_ops.all_ds_path)-2:
            all_ds_roi_counter.append(ds_roi_counter_mov)

        # save all to track ops
        track_ops.all_ds_all_roi_array_ref = all_ds_all_roi_array_ref
        track_ops.all_ds_all_roi_array_mov = all_ds_all_roi_array_mov
        track_ops.all_ds_all_roi_array_reg = all_ds_all_roi_array_reg
        track_ops.all_ds_roi_counter = all_ds_roi_counter

    return all_ds_all_roi_array_ref, all_ds_all_roi_array_mov, all_ds_all_roi_array_reg, all_ds_roi_counter



import os
from PyQt5.QtWidgets import QVBoxLayout, QWidget,  QHBoxLayout, QPushButton, QFileDialog, QLineEdit, QLabel, QFormLayout, QListWidget, QMessageBox,QListWidgetItem, QInputDialog,QCheckBox,QSizePolicy,QComboBox,QDialog
from PyQt5.QtCore import Qt
from track2p.t2p import run_t2p
from track2p.ops.default import DefaultTrackOps
from track2p.gui.custom_wd import CustomDialog
class Track2pWindow(QWidget):
        """it is used to set the parameters of the track2p algorithm"""
        def __init__(self, main_wd):
            super(Track2pWindow,self).__init__()
            self.main_window = main_wd
            layout = QFormLayout()
            self.setLayout(layout)
            self.track_ops = DefaultTrackOps()
            self.saved_directory=None
            self.plane=None
           
       
            instruction1=QLabel(" Import the directory containing subfolders of different recordings for the same subject:")
            self.import_recording_button = QPushButton("Import", self)
            self.import_recording_button.clicked.connect(self.import_path_to_recordings)
            layout.addRow(instruction1,self.import_recording_button)
            
            instruction2= QLabel("Here is the path of the imported directory:")
            instruction2.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            self.path_recording=QLabel()
            self.path_recording.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            layout.addRow(instruction2,self.path_recording)
            
            instruction3=QLabel("Once loaded press '->' to add to the list of paths to use for track2p (in the right order):")
            
            instruction4= QLabel("Choose an option for selecting suite2p ROIs:")
            field_checkbox= QVBoxLayout()
            self.checkbox1 = QCheckBox('manually curated', self)
            self.checkbox2 = QCheckBox('iscell threshold', self)
            self.checkbox2.stateChanged.connect(self.display_iscell)
            field_checkbox.addWidget(self.checkbox1)
            field_checkbox.addWidget(self.checkbox2)

            
            self.is_cell_thr=  QLineEdit()
            self.is_cell_thr.setVisible(False)
            self.is_cell_thr.setText('0.5')
            self.is_cell_thr.setFixedWidth(50)
             

            file_layout = QHBoxLayout()
       
            self.computer_file_list = QListWidget(self)
            self.computer_file_list.setFixedHeight(200) 
            self.move_to_computer_list = QPushButton("<-", self)
            self.move_to_computer_list.clicked.connect(self.move_file_to_computer_list)
            self.move_to_paths_list = QPushButton("->", self)
            self.move_to_paths_list.clicked.connect(self.move_file_to_paths_list)
            self.paths_list=QListWidget(self)
            self.paths_list.setFixedHeight(200) 
            
            file_layout.addWidget(self.computer_file_list)
            file_layout.addWidget(self.move_to_computer_list)
            file_layout.addWidget(self.move_to_paths_list)
            file_layout.addWidget(self.paths_list)
            layout.addRow(instruction3, file_layout)
            layout.addRow(instruction4,field_checkbox)
            layout.addRow("iscell threshold:",self.is_cell_thr) 
            

            instruction5=QLabel("Choose the channel to use for day-to-day registration (0 : functional, 1 : anatomical (if available))")
            self.reg_chan= QLineEdit()
            self.reg_chan.setFixedWidth(50)
            self.reg_chan.setText('0')
            layout.addRow(instruction5,self.reg_chan)

            trsfrm_type=QLabel("Choose the type of transformation to use for day-to-day registration:")
            self.trsfrm_type=QComboBox()
            self.trsfrm_type.addItem("affine")
            self.trsfrm_type.addItem("rigid")
            self.trsfrm_type.setCurrentIndex(0)
            layout.addRow(trsfrm_type,self.trsfrm_type)

            compute_iou=QLabel("iou_dist_thr:")
            self.compute_iou= QLineEdit()
            self.compute_iou.setFixedWidth(50)
            self.compute_iou.setText('16')
            layout.addRow(compute_iou,self.compute_iou)

            thr_method=QLabel("Thresholding method for IoU histogram:")
            self.thr_method=QComboBox()
            self.thr_method.addItem("min")
            self.thr_method.addItem("otsu")
            self.thr_method.setCurrentIndex(0)
            layout.addRow(thr_method,self.thr_method)
            
            instruction6=QLabel("Import the directory where the outputs will be saved (a 'track2p' sub-folder will be created here):")
            self.t2p_path_button = QPushButton("Import", self)
            self.t2p_path_button.clicked.connect(self.save_directoy)
            layout.addRow(instruction6,self.t2p_path_button)
            
            instruction7= QLabel("Here is the path to access the track2P folder:")
            instruction7.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            self.save_path=QLabel()
            self.save_path.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            layout.addRow(instruction7,self.save_path)
            
            instruction8= QLabel("Save the outputs in suite2p format (containing cells tracked on all days):")
            self.checkbox3 = QCheckBox(self)
            layout.addRow(instruction8,self.checkbox3)
            

            self.run_button = QPushButton("Run", self)
            self.run_button.clicked.connect(self.run)
            layout.addRow("Run the algorithm:", self.run_button)
            
            terminal_intruction=QLabel("To monitor the progress of the algorith see outputs in the terminal where the GUI was launched from")
            layout.addRow(terminal_intruction)
     
    
        def display_iscell(self,state):
            if state == Qt.Checked:
                self.is_cell_thr.setVisible(True)
            else:
                self.is_cell_thr.setVisible(False)

        def run(self):
    
            stored_all_ds_path = []
            for i in range(self.paths_list.count()):
                item=self.paths_list.item(i).data(Qt.UserRole)
                item_universel=item.replace("\\", "/")
                stored_all_ds_path.append(item_universel)
            self.track_ops.all_ds_path= stored_all_ds_path
            save_path=self.saved_directory
            save_path=save_path.replace("\\", "/")
            self.track_ops.save_path = save_path
            self.track_ops.reg_chan=int(self.reg_chan.text())
            self.track_ops.transform_type=self.trsfrm_type.currentText()
            self.track_ops.iou_dist_thr=int(self.compute_iou.text())
            self.track_ops.thr_method=self.thr_method.currentText()
            print("transformation type:", self.track_ops.transform_type)
            print("iou_dist_thr:", self.track_ops.iou_dist_thr)
            print("thr_method:", self.track_ops.thr_method)
            if self.checkbox1.isChecked():
                self.track_ops.iscell_thr=None
            if self.checkbox2.isChecked():
                self.track_ops.iscell_thr=float(self.is_cell_thr.text())
            if self.checkbox3.isChecked():
                self.track_ops.save_in_s2p_format=True
            print("All parameters have been recorded ! The track2p algorithm is running...")
            run_t2p(self.track_ops)
            self.open_track2p_in_gui()

    
        def open_track2p_in_gui(self):
            reply = QMessageBox.question(self, "", "Run completed successfully!\nDo you want to launch the gui?", QMessageBox.Yes | QMessageBox.No, QMessageBox.No)

            if reply == QMessageBox.Yes:
                #print("Opening GUI...")
                dialog = CustomDialog(self.main_window, self.saved_directory)
                dialog.exec_()
            if reply == QMessageBox.No:
                pass
        
        def save_directoy(self):
            saved_directory= QFileDialog.getExistingDirectory(self, "Select Directory")
            if saved_directory:
                self.saved_directory=saved_directory
                self.save_path.setText(f'{self.saved_directory}')
                
        def import_path_to_recordings(self):
            directory = QFileDialog.getExistingDirectory(self, "Select Directory")
            if directory:
                self.path_recording.setText(f'{directory}')
                self.computer_file_list.clear()
                files= os.listdir(directory)
                for file in sorted(files):
                    full_path = os.path.join(directory, file)
                    item=QListWidgetItem(file)
                    item.setData(Qt.UserRole, full_path)
                    self.computer_file_list.addItem(item)


        def move_file_to_paths_list(self):
            selected_items = self.computer_file_list.selectedItems()
            for item in selected_items:
                self.computer_file_list.takeItem(self.computer_file_list.row(item))
                self.paths_list.addItem(item)
     

        def move_file_to_computer_list(self):
            selected_items = self.paths_list.selectedItems()
            for item in selected_items:
                self.paths_list.takeItem(self.paths_list.row(item))
                self.computer_file_list.addItem(item)
    






from PyQt5.QtWidgets import  QWidget, QPushButton, QFileDialog, QLineEdit, QLabel, QFormLayout,QComboBox

class ImportWindow(QWidget):
        
        def __init__(self, main_wd):
            super(ImportWindow,self).__init__()
            
            self.main_window = main_wd
            self.path_to_t2p = None
            self.plane=None
            
            layout = QFormLayout()
            
            import_label=QLabel("Import the directory containing the track2p folder:")
            self.import_button = QPushButton("Import", self)
            self.import_button.clicked.connect(self.save_t2p_path)
            layout.addRow(import_label,self.import_button)
            path_label=QLabel("Here is the path of the imported directory:")
            self.path = QLabel()
            layout.addRow(path_label,self.path)
            plane_label= QLabel("Choose the plane to analyze:")
            self.textbox = QLineEdit(self)
            self.textbox.setFixedWidth(50) 
            self.textbox.setText('0')
            trace_label=QLabel("Choose trace type:")
            self.trace_choice=QComboBox()
            self.trace_choice.addItem("F")
            self.trace_choice.addItem("dF/F0")
            self.trace_choice.addItem("spks")
            channel_label=QLabel("Choose the channel to show in the mean images:")
            self.channel_choice=QComboBox()
            self.channel_choice.addItem("0")
            self.channel_choice.addItem("1")
            self.channel_choice.addItem("Vcorr")
            self.channel_choice.addItem("max_proj")
            layout.addRow(plane_label,self.textbox)
            layout.addRow(channel_label,self.channel_choice)
            layout.addRow(trace_label,self.trace_choice)
            label_run= QLabel("Run the analysis:")
            self.run_button = QPushButton("Run", self)
            self.run_button.clicked.connect(self.run)
            layout.addRow(label_run,self.run_button)
            
            self.setLayout(layout)
            
                
        def save_t2p_path(self):
            path_to_t2p= QFileDialog.getExistingDirectory(self, "Select Directory")
            if path_to_t2p:
                self.path_to_t2p=path_to_t2p
                self.path.setText(f'{self.path_to_t2p}')  
                
        def run(self):
            self.plane = int(self.textbox.text())
            self.trace_type=self.trace_choice.currentText()
            self.main_window.central_widget.data_management.import_files(self.path_to_t2p, plane=self.plane, trace_type=self.trace_type, channel=self.channel_choice.currentText())
                


import os
from PyQt5.QtWidgets import QVBoxLayout, QWidget,  QHBoxLayout, QPushButton, QFileDialog, QLineEdit, QLabel, QFormLayout, QListWidget, QMessageBox,QListWidgetItem, QInputDialog,QCheckBox,QSizePolicy,QComboBox,QDialog
from PyQt5.QtCore import Qt
from track2p.t2p import run_t2p
from track2p.ops.default import DefaultTrackOps

class CustomDialog(QDialog):
    def __init__(self, main_window, save_directory):
        super(CustomDialog,self).__init__()
        self.main_window = main_window
        self.save_directory = save_directory
        self.cancel_clicked = False

        # Layout
        layout = QVBoxLayout(self)
        
        # Plane input
        self.plane_label = QLabel("Enter your plane:", self)
        layout.addWidget(self.plane_label)
        self.plane_input = QLineEdit(self)
        layout.addWidget(self.plane_input)
        
        # Trace type input
        self.trace_label = QLabel("Choose trace type:", self)
        layout.addWidget(self.trace_label)
        self.trace_combo = QComboBox(self)
        self.trace_combo.addItems(["F", "spks", "dF/F0"])  
        layout.addWidget(self.trace_combo)
        
        # OK and Cancel buttons
        self.ok_button = QPushButton("OK", self)
        self.ok_button.clicked.connect(self.import_data)
        layout.addWidget(self.ok_button)
        
        self.cancel_button = QPushButton("Cancel", self)
        self.cancel_button.clicked.connect(self.on_cancel_clicked)
        layout.addWidget(self.cancel_button)

    def get_inputs(self):
        return self.plane_input.text(), self.trace_combo.currentText()
    
    def on_cancel_clicked(self):
        self.cancel_clicked = True
        self.import_data()

    def import_data(self):
        if self.cancel_clicked:
            self.close()
            pass
        else: 
            print("Opening GUI...")
            plane_text, trace_type = self.get_inputs()
            self.plane = int(plane_text)
            self.trace_type = trace_type
            print(f"Converted plane: {self.plane}, Trace type: {self.trace_type}")  # Debugging print
            self.main_window.central_widget.data_management.import_files(t2p_folder_path = self.save_directory, plane=self.plane, trace_type=self.trace_type)
            self.close()




import os

# make a directory based on path if it doesn't exist yet

def make_dir(path):
    if not os.path.exists(path):
        os.makedirs(path)
        print('Created directory: ' + path)
    else:
        print('Directory already exists: ' + path)


from setuptools import find_packages, setup

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='track2p',
    version='0.5.8',
    packages=find_packages(),
    install_requires=[
        'numpy==1.23.5',
        'matplotlib==3.5.3',
        'scikit-image==0.20.0',
        'itk==5.4rc2',
        'PyQt5==5.15.10',
        'qtpy==2.4.1',
        'tqdm==4.66.2',
        'scikit-learn==1.4.0',
        'openTSNE==1.0.1',
        'pandas==1.5.3',
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    package_data={
        '': ['resources/logo.png'],
    },
)


import colorsys
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import zscore
from PyQt5.QtCore import Qt
import matplotlib.patches as patches
from PyQt5 import QtCore
from scipy.ndimage import maximum_filter1d, minimum_filter1d, gaussian_filter


class FluorescencePlotWidget(FigureCanvas):
    """this class is used to display the fluorescence of the selected cell across days. It also allows to select a region of interest (ROI) on the fluorescence plot and zoom in on the selected ROI"""
    def __init__(self, all_f_t2p=None, all_ops=None, colors=None, all_stat_t2p=None):
        self.fig, self.ax_fluorescence = plt.subplots(1, 1)
        super().__init__(self.fig)
        self.all_f_t2p = all_f_t2p
        self.all_ops=all_ops
        self.fig.set_facecolor('black')
        self.colors = colors
        self.all_stat_t2p= all_stat_t2p
  
        
        self.rect = patches.Rectangle((0,0), 1, 1, color='white', linewidth=2) 
        
        self.fig.canvas.mpl_connect('button_press_event', self.on_press)
        self.fig.canvas.mpl_connect('button_release_event', self.on_release)
        self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.fig.canvas.mpl_connect('key_press_event', self.on_enter_pressed)
        self.setFocusPolicy(Qt.StrongFocus)
        self.setFocus()
        
        self.initial_xlim = None
        self.initial_ylim = None 
        self.cmd_pressed = False
        
    def draw_rectangle(self):
        """it draws a rectangle on the fluorescence plot"""
        #if the rectangle is already on the plot, it is removed
        if self.rect in self.ax_fluorescence.patches:
            self.ax_fluorescence.patches.remove(self.rect)
        self.rect = patches.Rectangle((self.x0, self.y0), self.x1 - self.x0, self.y1 - self.y0, fill=None, linewidth=2, edgecolor='white') #create a rectangle, fill is set to None to make the rectangle transparent
        self.rect.set_zorder(10)
        self.ax_fluorescence.add_patch(self.rect)
        self.draw()
        
    def on_key_press(self, event):
        """it allows to reset the zoom of the fluorescence plot. It is called when the Command key and - key are pressed."""
        if event.key == 'r':
            self.ax_fluorescence.set_xlim(self.initial_xlim)
            self.ax_fluorescence.set_ylim(self.initial_ylim)
            self.fig.canvas.draw()
    
    def draw_point(self):
        if hasattr(self, 'point') and self.point in self.ax_fluorescence.collections:
            print(self.ax_fluorescence.collections)
            self.ax_fluorescence.collections.remove(self.point)
        self.point = self.ax_fluorescence.scatter([self.x0], [self.y0], s=5, color='w')  # Create a new point
        self.draw()
  
    def on_press(self, event):
        """it allows to draw a point on the fluorescence plot. It is called when the mouse button is pressed."""
        self.x0 = event.xdata #x coordinate of the mouse cursor
        self.y0 = event.ydata  #y coordinate of the mouse cursor
        self.draw_point()

    def on_release(self, event):
        """It allows to draw a rectangle on the fluorescence plot. It is called when the mouse button is released."""
        self.x1 = event.xdata 
        self.y1 = event.ydata
        self.rect.set_width(self.x1 - self.x0) 
        self.rect.set_height(self.y1 - self.y0)
        self.rect.set_xy((self.x0, self.y0)) #set the position of the rectangle
        self.draw_rectangle()
        
        
    def on_enter_pressed(self, event):
        """it allows to zoom in of the fluorescence plot. It is called when the Command key and + key are pressed."""
        if event.key == 'enter':
            self.ax_fluorescence.set_xlim(self.x0, self.x1)
            self.ax_fluorescence.set_ylim(self.y0, self.y1)
            self.rect.set_visible(False) 
            self.point.set_visible(False)
            self.fig.canvas.draw()


    def display_all_f_t2p(self, selected_cell_index):
        """it plots the fluroescence of the selected cell across days where each curve being a different day (the curve at the top of the plot is the first day)"""
        
        if self.all_f_t2p is not None and selected_cell_index is not None:
            self.ax_fluorescence.clear()
            self.ax_fluorescence.set_facecolor('black')
            self.ax_fluorescence.tick_params(axis='x', colors='white') 
            self.ax_fluorescence.tick_params(axis='y', colors='white')  
            self.ax_fluorescence.xaxis.label.set_color('white') 
            self.ax_fluorescence.yaxis.label.set_color('white')
            self.ax_fluorescence.spines['bottom'].set_color('#666') 
        
            
            for i, fluorescence_data in list(enumerate(reversed(self.all_f_t2p))):
                fluorescence_zscore = zscore(fluorescence_data, axis=1, ddof=1) #zscore is used to normalize the fluorescence data
                offset = i * 12 #
                y_values = fluorescence_zscore[selected_cell_index, :] + offset 
                color = self.colors[selected_cell_index] 
                #create a gradient of colors for the curves (the darkest shade is for the last day and the lightest shade is for the first day)
                if i == 0:
                    color = color 
                else:
                    h, l, s = colorsys.rgb_to_hls(*color)  
                    l_range = 1 - (l + 0.05)  
                    l_add = l_range/len(self.all_f_t2p) 
                    
                    adjusted_luminosity = l + (l_add *i) 
                    color = colorsys.hls_to_rgb(h, adjusted_luminosity, s)

                ops=self.all_ops[i]
                fs = ops['fs']
                tstamps = np.arange(len(y_values))
                
                if len(tstamps) != len(y_values):
                    raise ValueError(f"tstamps and y_values must have the same length, but have lengths {len(tstamps)} and {len(y_values)}")
                self.ax_fluorescence.plot(tstamps,y_values, label=f'Curve {i + 1}', color= color)
            
            self.ax_fluorescence.set_xticks([0, int(len(tstamps)/2), len(tstamps)])
            self.ax_fluorescence.set_yticklabels([])
            self.ax_fluorescence.get_yaxis().set_visible(False)
            self.ax_fluorescence.set_xlabel('Frames')
            self.initial_xlim=self.ax_fluorescence.get_xlim()
            self.initial_ylim=self.ax_fluorescence.get_ylim()
            self.fig.tight_layout() 
            self.draw()
            
         
        pass

    



from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import numpy as np
import matplotlib.pyplot as plt
import skimage


class ZoomPlotWidget(FigureCanvas):
    """It is used to display the roi of the selected cell across days with each zoom being a different day. 
    The roi is centered on the median coordinates of the selected cell. The mean image of the recording is used to create the zooms. 
    The probability of the cell being a cell and the index of the cell in the suite2p files associated with each recording (day) are also displayed under the zooms."""
    
    def __init__(self,  all_ops=None, all_stat_t2p=None, colors=None, all_iscell_t2p=None,t2p_match_mat_allday =None,track_ops=None, imgs=None):
        nb_plot=len(all_iscell_t2p)
        self.fig, self.ax_zoom = plt.subplots(1, nb_plot, figsize = (10*nb_plot, nb_plot), gridspec_kw={'width_ratios': [1] * nb_plot},facecolor='black')
        super().__init__(self.fig)
        self.track_ops=track_ops
        self.all_ops = all_ops
        self.all_stat_t2p = all_stat_t2p
        self.colors = colors
        self.all_is_cell=all_iscell_t2p
        self.t2p_match_mat_allday =t2p_match_mat_allday 
        self.imgs= imgs
        self.coord_dict={} 

    def display_zooms(self, selected_cell_index):
        """It is called when the application is opened and a cell is selected."""
#it is used to store the roi and the median coordinates of the selected cell for each recording (day)
        self.coord_dict={}
        if self.all_ops is not None and self.all_stat_t2p is not None and self.all_is_cell is not None:
            for i in range(len(self.imgs)):
                wind = 20
                #mean_img = self.all_ops[i]['meanImg']
                mean_img=self.imgs[i]
                stat_t2p = self.all_stat_t2p[i]
                median_coord = stat_t2p[selected_cell_index]['med']
                print(f'median_coord : ', median_coord)

                print(mean_img.shape)
                # Définir la taille de la marge
                margin = 20

                # Obtenir les dimensions de l'image originale
                Ly, Lx = mean_img.shape

                # Créer une nouvelle image avec des dimensions augmentées
                new_Ly = Ly + 2 * margin
                new_Lx = Lx + 2 * margin
                range_img = np.zeros((new_Ly, new_Lx))
             
                
                # Copier mean_img au centre de la nouvelle image
                range_img[margin:margin + Ly, margin:margin + Lx] = mean_img
                
                # Ajouter la marge aux coordonnées médianes
                median_x = int(median_coord[1]) + margin
                median_y = int(median_coord[0]) + margin

                # Calculer les nouvelles coordonnées de la ROI
                x_start = median_x - wind
                x_end = median_x + wind
                y_start = median_y - wind
                y_end = median_y + wind

                # Extraire la ROI de l'image avec la marge
                roi = range_img[y_start:y_end, x_start:x_end]

                print(range_img.shape)
                print(roi.shape)

            
                #prob and index 
                iscell=self.all_is_cell[i]
                if self.track_ops.iscell_thr==None:
                    indices_lignes_1 = np.where(iscell[:,0]==1)[0]
                    match_index=self.t2p_match_mat_allday[selected_cell_index,i]
                    true_index=indices_lignes_1[match_index]
                else:
                    indices_lignes_1= np.where(iscell[:,1]>self.track_ops.iscell_thr)[0]
                    match_index=self.t2p_match_mat_allday[selected_cell_index,i]
                    true_index=indices_lignes_1[match_index]
                prob=round(iscell[true_index,1],2)

                ypix=stat_t2p[selected_cell_index]['ypix']
                xpix=stat_t2p[selected_cell_index]['xpix']


                ax = self.ax_zoom[i]
                color = self.colors[selected_cell_index]

                mask=np.zeros((2*wind,2*wind))
                mask[ypix-median_coord[0]+wind,xpix-median_coord[1]+wind]=1
                ax.clear()
                ax.contour(mask,levels=[0.5], colors=[color],linewidths=2)

                #last_img=list(self.coord_dict.values())[-1][0]
                #match_roi=skimage.exposure.match_histograms(list[0], last_img, channel_axis=None)
                
                ax.imshow(roi, cmap='gray')
                ax.set_title(f'Day {i + 1}', color='white', fontsize=10)
                ax.text(0.5, -0.2, f'i: {true_index}', color='white', fontsize=10, ha='center', va='center', transform=ax.transAxes)
                ax.text(0.5, -0.4, f'p: {prob}', color='white', fontsize=10, ha='center', va='center', transform=ax.transAxes)

                ax.axis('off')
        
        self.draw()
                    


import numpy as np

def get_all_ds_img_for_reg(all_ds_avg_ch1, all_ds_avg_ch2, track_ops): # chooses which channel to use for registration and returns all_ref_img and all_mov_img (shifted by one day to always register to previous day)
    if track_ops.reg_chan==0:
        all_ds_avg = all_ds_avg_ch1
    elif track_ops.reg_chan==1:
        all_ds_avg = all_ds_avg_ch2
        print('WARNING: using anatomical channel for registration (this is not always available)')

    all_ds_ref_img = []
    all_ds_mov_img = []

    for i in range(len(track_ops.all_ds_path)-1):
        ds_ref_img = []
        ds_mov_img = []
        
        for j in range(track_ops.nplanes):
            ds_ref_img.append(all_ds_avg[i][j])
            ds_mov_img.append(all_ds_avg[i+1][j])

        all_ds_ref_img.append(ds_ref_img)
        all_ds_mov_img.append(ds_mov_img)

    track_ops.all_ds_ref_img = all_ds_ref_img
    track_ops.all_ds_mov_img = all_ds_mov_img

    return all_ds_ref_img, all_ds_mov_img


def get_ref_reg_inters(all_roi_array_ref, all_roi_array_nonref):
    # get the projection of all rois
    all_roi_array_ref_proj = np.sum(all_roi_array_ref, axis=2) > 0
    all_roi_array_nonref_proj = np.sum(all_roi_array_nonref, axis=2) > 0

    # now get the intersection of the reg and ref   
    all_roi_array_inters = np.logical_and(all_roi_array_ref_proj, all_roi_array_nonref_proj)

    # make rgb image of intersection
    ref_reg_inters = np.ones((all_roi_array_inters.shape[0], all_roi_array_inters.shape[1], 3))
    ref_reg_inters[:,:,0] 
    ref_reg_inters[:,:,1] -= all_roi_array_inters/6 # orange tint
    ref_reg_inters[:,:,2] -= all_roi_array_inters

    return ref_reg_inters

def get_all_ref_nonref_inters(all_ds_all_roi_array_ref, all_ds_all_roi_array_nonref, track_ops):
    all_ds_all_ref_nonref_inters = []
    for i in range(len(track_ops.all_ds_path)-1):
        ds_all_ref_nonref_inters = []
        for j in range(track_ops.nplanes):
            all_roi_array_ref = all_ds_all_roi_array_ref[i][j]
            all_roi_array_nonref = all_ds_all_roi_array_nonref[i][j]
            ref_nonref_inters = get_ref_reg_inters(all_roi_array_ref, all_roi_array_nonref)
            ds_all_ref_nonref_inters.append(ref_nonref_inters)
        all_ds_all_ref_nonref_inters.append(ds_all_ref_nonref_inters)
    return all_ds_all_ref_nonref_inters


from PyQt5.QtWidgets import QStatusBar,QWidget,  QHBoxLayout, QSpinBox,QPushButton,QLabel
import numpy as np 
import os 

class StatusBar(QStatusBar):
    def __init__(self, main_window):
        super().__init__()

        self.main_window = main_window
        self.central_widget = self.main_window.central_widget
        self.vector_curation_t2p=None 
        self.init_status_bar()

    def init_status_bar(self):
        status_widget = QWidget()
        layout = QHBoxLayout()

        self.spin_box = QSpinBox()
        self.spin_box.setFixedWidth(100)
        self.spin_box.valueChanged.connect(self.iterate_all_rois)

        self.roi_state = QLabel("state of ROI: ")
        self.roi_state.setFixedWidth(100)
        self.roi_state_value = QLabel()
        self.roi_state_value.setFixedWidth(30)

        not_cell_button = QPushButton('✖️')
        not_cell_button.setFixedSize(20, 20)
        not_cell_button.setStyleSheet("background-color: red; color: white; border: none;")
        not_cell_button.clicked.connect(self.set_roi_as_not_cell)

        cell_button = QPushButton('✓')
        cell_button.setFixedSize(20, 20)
        cell_button.setStyleSheet("background-color: green; color: white; border: none;")
        cell_button.clicked.connect(self.set_roi_as_cell)

        reset_button = QPushButton('Apply curation')
        reset_button.setFixedSize(100, 20)
        reset_button.setStyleSheet("background-color: grey; color: white; border: none;")
        reset_button.clicked.connect(self.main_window.central_widget.create_mean_img_from_curation)


        layout.addWidget(self.spin_box)
        layout.addWidget(self.roi_state)
        layout.addWidget(self.roi_state_value)
        layout.addWidget(not_cell_button)
        layout.addWidget(cell_button)
        layout.addWidget(reset_button)

        status_widget.setLayout(layout)
        self.addWidget(status_widget)
        
    def iterate_all_rois(self):
        current_ROI = self.spin_box.value()
        value=self.vector_curation_t2p[current_ROI] 
        self.roi_state_value.setText(f"{value}") 
        self.central_widget.update_selection(current_ROI) 

    def set_roi_as_not_cell(self):
        plane=self.central_widget.data_management.plane
        key='vector_curation_plane_' + str(plane)
        self.vector_curation_t2p = self.main_window.central_widget.data_management.vector_curation_t2p
        if self.vector_curation_t2p[self.spin_box.value()] ==1:
            self.vector_curation_t2p[self.spin_box.value()]= 0
            current_ROI = self.spin_box.value() 
            value=self.vector_curation_t2p [current_ROI] 
            self.roi_state_value.setText(f"{value}")
            self.central_widget.track_ops_dict[key] = self.vector_curation_t2p
            np.save(os.path.join(self.central_widget.data_management.track_ops.save_path, "track_ops.npy"), self.central_widget.track_ops_dict)
           
    def set_roi_as_cell(self):
        plane=self.central_widget.data_management.plane
        key='vector_curation_plane_' + str(plane)
        self.vector_curation_t2p = self.main_window.central_widget.data_management.vector_curation_t2p
        if self.vector_curation_t2p[self.spin_box.value()] ==0:
            self.vector_curation_t2p[self.spin_box.value()]= 1
            current_ROI = self.spin_box.value() 
            value=self.vector_curation_t2p[current_ROI] 
            self.roi_state_value.setText(f"{value}")
            self.central_widget.track_ops_dict[key] = self.vector_curation_t2p
            np.save(os.path.join(self.central_widget.data_management.track_ops.save_path, "track_ops.npy"), self.central_widget.track_ops_dict)


    
    


import numpy as np
import os

from track2p.io.utils import make_dir

def save_track_ops(track_ops):
    # remove attributes taking a lot of memory (e.g. rois etc.)
    del track_ops.all_ds_all_roi_array_mov
    del track_ops.all_ds_all_roi_array_ref
    del track_ops.all_ds_all_roi_array_reg

    track_ops_dict = track_ops.to_dict() # convert to dictionary for compatibility
    
    np.save(os.path.join(track_ops.save_path, 'track_ops.npy'), track_ops_dict, allow_pickle=True)
    print('Saved track_ops.npy in ' + track_ops.save_path)

def save_all_pl_match_mat(all_pl_match_mat, track_ops):
    for (i, all_pl_match_mat) in enumerate(all_pl_match_mat):
        np.save(os.path.join(track_ops.save_path, f'plane{i}_match_mat.npy'), all_pl_match_mat, allow_pickle=True)